from typing import Optional, Dict, Any
import inspect

import sqlalchemy as sa  # noqa: I251
from sqlalchemy.sql import sqltypes

from dlt.common import json
from dlt.common.exceptions import TerminalValueError
from dlt.common.typing import TLoaderFileFormat
from dlt.common.destination.capabilities import DataTypeMapper, DestinationCapabilitiesContext
from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.schema.typing import TColumnSchema


# TODO: base type mapper should be a generic class to support TypeEngine instead of str types
class SqlalchemyTypeMapper(DataTypeMapper):
    """Maps dlt data types to SQLAlchemy column types and back.

    to_destination_type dispatches to per-type visitor methods named
    _db_type_from_<dlt_type>_type (e.g. _db_type_from_json_type). Subclasses
    can override individual visitors to customize mapping for a single dlt type
    without reimplementing the full dispatch. Overriding to_destination_type
    directly is still supported for backward compatibility.
    """

    def __init__(
        self,
        capabilities: DestinationCapabilitiesContext,
        dialect: Optional[sa.engine.Dialect] = None,
    ):
        super().__init__(capabilities)
        # Mapper is used to verify supported types without client, dialect is not important for this
        self.dialect = dialect or sa.engine.default.DefaultDialect()

    def _db_integer_type(self, precision: Optional[int]) -> sa.types.TypeEngine:
        if precision is None:
            return sa.BigInteger()
        elif precision <= 16:
            return sa.SmallInteger()
        elif precision <= 32:
            return sa.Integer()
        elif precision <= 64:
            return sa.BigInteger()
        raise TerminalValueError(f"Unsupported `{precision=:}` for integer type.")

    def _create_date_time_type(
        self, sc_t: str, precision: Optional[int], timezone: Optional[bool]
    ) -> sa.types.TypeEngine:
        """Use the dialect specific datetime/time type if possible since the generic type doesn't accept precision argument"""
        precision = precision if precision is not None else self.capabilities.timestamp_precision
        base_type: sa.types.TypeEngine
        timezone = timezone is None or bool(timezone)
        if sc_t == "timestamp":
            base_type = sa.DateTime()
        elif sc_t == "time":
            base_type = sa.Time()

        dialect_type = type(
            self.dialect.type_descriptor(base_type)
        )  # Get the dialect specific subtype

        # Find out whether the dialect type accepts precision or fsp argument
        params = inspect.signature(dialect_type).parameters
        kwargs: Dict[str, Any] = dict(timezone=timezone)
        if "fsp" in params:
            kwargs["fsp"] = precision  # MySQL uses fsp for fractional seconds
        elif "precision" in params:
            kwargs["precision"] = precision
        return dialect_type(**kwargs)  # type: ignore[no-any-return,misc]

    def _create_double_type(self) -> sa.types.TypeEngine:
        if dbl := getattr(sa, "Double", None):
            # Sqlalchemy 2 has generic double type
            return dbl()  # type: ignore[no-any-return]
        return sa.Float(precision=53)  # Otherwise use float

    def _to_db_decimal_type(self, column: TColumnSchema) -> sa.types.TypeEngine:
        precision, scale = column.get("precision"), column.get("scale")
        if precision is None and scale is None:
            precision, scale = self.capabilities.decimal_precision
        return sa.Numeric(precision, scale)

    def db_type_from_text_type(
        self, column: TColumnSchema, table: PreparedTableSchema
    ) -> sqltypes.TypeEngine:
        """Converts dlt "text" — variable-length string, optional precision sets max length."""
        precision = column.get("precision")
        length = precision
        if length is None and column.get("unique"):
            length = 128
        if length is None:
            return sa.Text()
        else:
            return sa.String(length=length)

    def db_type_from_double_type(
        self, column: TColumnSchema, table: PreparedTableSchema
    ) -> sqltypes.TypeEngine:
        """Converts dlt "double" — 64-bit IEEE 754 floating point."""
        return self._create_double_type()

    def db_type_from_bool_type(
        self, column: TColumnSchema, table: PreparedTableSchema
    ) -> sqltypes.TypeEngine:
        """Converts dlt "bool" — true/false."""
        return sa.Boolean()

    def db_type_from_timestamp_type(
        self, column: TColumnSchema, table: PreparedTableSchema
    ) -> sqltypes.TypeEngine:
        """Converts dlt "timestamp" — date and time with optional timezone and fractional seconds."""
        return self._create_date_time_type(
            "timestamp", column.get("precision"), column.get("timezone")
        )

    def db_type_from_bigint_type(
        self, column: TColumnSchema, table: PreparedTableSchema
    ) -> sqltypes.TypeEngine:
        """Converts dlt "bigint" — integer with optional precision (16, 32, 64 bits)."""
        return self._db_integer_type(column.get("precision"))

    def db_type_from_binary_type(
        self, column: TColumnSchema, table: PreparedTableSchema
    ) -> sqltypes.TypeEngine:
        """Converts dlt "binary" — raw bytes, optional precision sets max length."""
        return sa.LargeBinary(length=column.get("precision"))

    def db_type_from_json_type(
        self, column: TColumnSchema, table: PreparedTableSchema
    ) -> sqltypes.TypeEngine:
        """Converts dlt "json" — arbitrary nested JSON data."""
        return sa.JSON(none_as_null=True)

    def db_type_from_decimal_type(
        self, column: TColumnSchema, table: PreparedTableSchema
    ) -> sqltypes.TypeEngine:
        """Converts dlt "decimal" — arbitrary-precision decimal number."""
        return self._to_db_decimal_type(column)

    def db_type_from_wei_type(
        self, column: TColumnSchema, table: PreparedTableSchema
    ) -> sqltypes.TypeEngine:
        """Converts dlt "wei" — large-precision integer for blockchain wei values."""
        wei_precision, wei_scale = self.capabilities.wei_precision
        return sa.Numeric(precision=wei_precision, scale=wei_scale)

    def db_type_from_date_type(
        self, column: TColumnSchema, table: PreparedTableSchema
    ) -> sqltypes.TypeEngine:
        """Converts dlt "date" — calendar date without time component."""
        return sa.Date()

    def db_type_from_time_type(
        self, column: TColumnSchema, table: PreparedTableSchema
    ) -> sqltypes.TypeEngine:
        """Converts dlt "time" — time of day with optional timezone and fractional seconds."""
        return self._create_date_time_type("time", column.get("precision"), column.get("timezone"))

    def to_destination_type(  # type: ignore[override]
        self, column: TColumnSchema, table: PreparedTableSchema = None
    ) -> sqltypes.TypeEngine:
        sc_t = column["data_type"]
        method = getattr(self, f"db_type_from_{sc_t}_type", None)
        if method is not None:
            return method(column, table)  # type: ignore[no-any-return]
        raise TerminalValueError(f"Unsupported data type: `{sc_t}`")

    def _from_db_integer_type(self, db_type: sa.Integer) -> TColumnSchema:
        if isinstance(db_type, sa.SmallInteger):
            return dict(data_type="bigint", precision=16)
        elif isinstance(db_type, sa.Integer):
            return dict(data_type="bigint", precision=32)
        elif isinstance(db_type, sa.BigInteger):
            return dict(data_type="bigint")
        return dict(data_type="bigint")

    def _from_db_decimal_type(self, db_type: sa.Numeric) -> TColumnSchema:
        precision, scale = db_type.precision, db_type.scale
        if (precision, scale) == self.capabilities.wei_precision:
            return dict(data_type="wei")

        return dict(data_type="decimal", precision=precision, scale=scale)

    def from_destination_type(  # type: ignore[override]
        self,
        db_type: sqltypes.TypeEngine,
        precision: Optional[int] = None,
        scale: Optional[int] = None,
    ) -> TColumnSchema:
        # TODO: pass the sqla type through dialect.type_descriptor before instance check
        # Possibly need to check both dialect specific and generic types
        if isinstance(db_type, sa.String):
            return dict(data_type="text")
        elif isinstance(db_type, sa.Float):
            return dict(data_type="double")
        elif isinstance(db_type, sa.Boolean):
            return dict(data_type="bool")
        elif isinstance(db_type, sa.DateTime):
            return dict(data_type="timestamp", timezone=db_type.timezone)
        elif isinstance(db_type, sa.Integer):
            return self._from_db_integer_type(db_type)
        elif isinstance(db_type, sqltypes._Binary):
            return dict(data_type="binary", precision=db_type.length)
        elif isinstance(db_type, sa.JSON):
            return dict(data_type="json")
        elif isinstance(db_type, sa.Numeric):
            return self._from_db_decimal_type(db_type)
        elif isinstance(db_type, sa.Date):
            return dict(data_type="date")
        elif isinstance(db_type, sa.Time):
            return dict(data_type="time")
        raise TerminalValueError(f"Unsupported db type: `{db_type}`")

    def ensure_supported_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema,
        loader_file_format: TLoaderFileFormat,
    ) -> None:
        pass


class MssqlVariantTypeMapper(SqlalchemyTypeMapper):
    def db_type_from_text_type(
        self, column: TColumnSchema, table: PreparedTableSchema
    ) -> sqltypes.TypeEngine:
        dt = super().db_type_from_text_type(column, table)
        precision = column.get("precision")
        length = precision
        if length is None and column.get("unique"):
            length = 4000  # max regular varchar
        if length is None:
            return dt.with_variant(sa.UnicodeText(), "mssql")  # type: ignore[no-any-return]
        else:
            return dt.with_variant(sa.Unicode(length=length), "mssql")  # type: ignore[no-any-return]


class MysqlVariantTypeMapper(SqlalchemyTypeMapper):
    def db_type_from_double_type(
        self, column: TColumnSchema, table: PreparedTableSchema
    ) -> sqltypes.TypeEngine:
        from sqlalchemy.dialects.mysql import DOUBLE

        return DOUBLE()

    def db_type_from_timestamp_type(
        self, column: TColumnSchema, table: PreparedTableSchema
    ) -> sqltypes.TypeEngine:
        # Special case, type_descriptor does not return the specific datetime type
        from sqlalchemy.dialects.mysql import DATETIME

        precision = column.get("precision")
        precision = precision if precision is not None else self.capabilities.timestamp_precision
        return DATETIME(fsp=precision)


class HexVarBinary(sa.TypeDecorator):
    impl = sa.String

    def bind_processor(self, dialect: sa.engine.Dialect) -> Any:
        import binascii

        return lambda v: None if v is None else binascii.hexlify(v).decode()


class JSONString(sa.TypeDecorator):
    """Custom SQLAlchemy type that stores JSON data as a string in the database.

    Automatically serializes Python objects to JSON strings on write and
    deserializes JSON strings back to Python objects on read.
    """

    impl = sa.String
    cache_ok = True

    def process_bind_param(self, value: Any, dialect: sa.engine.Dialect) -> Any:
        if value is None:
            return None

        if isinstance(value, str):
            return value

        return json.dumps(value)

    def process_result_value(self, value: Any, dialect: sa.engine.Dialect) -> Any:
        if value is None:
            return None

        return json.loads(value)


class TrinoVariantTypeMapper(SqlalchemyTypeMapper):
    def db_type_from_binary_type(
        self, column: TColumnSchema, table: PreparedTableSchema
    ) -> sqltypes.TypeEngine:
        dt = super().db_type_from_binary_type(column, table)
        return dt.with_variant(HexVarBinary(), "trino")  # type: ignore[no-any-return]

    def db_type_from_json_type(
        self, column: TColumnSchema, table: PreparedTableSchema
    ) -> sqltypes.TypeEngine:
        return JSONString()



from typing import Any, Dict, Optional
from dlt.common.configuration.specs.base_configuration import (
    CredentialsConfiguration,
    configspec,
)


@configspec
class HfCredentials(CredentialsConfiguration):
    """Credentials for HF filesystem, compatible with fsspec HF protocol at huggingface_hub.HfFileSystem.

    Authentication is attempted in the following order of priority:
        - the `hf_token` argument
        - the `HF_TOKEN` environment variable
        - a local token created with the `hf auth login` command

    For more information about parameters:
    https://huggingface.co/docs/huggingface_hub/en/quick-start#login-command
    """

    hf_token: Optional[str] = None
    """Hugging Face authentication token. Leave empty to attempt authentication through `HF_TOKEN` environment variable or locally saved token."""
    hf_endpoint: Optional[str] = None
    """Hugging Face API endpoint. Leave empty to use the default https://huggingface.co."""

    def to_hffs_credentials(self) -> Dict[str, Any]:
        """Return a dict that can be passed to fsspec huggingface_hub.HfFileSystem."""
        return {
            "token": self.hf_token,
            "endpoint": self.hf_endpoint,
        }

    def to_hf_api_credentials(self) -> Dict[str, Any]:
        """Return a dict that can be passed to huggingface_hub.HfApi."""
        return {
            "token": self.hf_token,
            "endpoint": self.hf_endpoint,
        }



import os
from pathlib import Path
import yaml
from typing import Any, Dict, List, Sequence, Tuple, cast
from inspect import signature
import dlt

from dlt.common.json import json
from dlt.common.pendulum import pendulum
from dlt.common.pipeline import get_dlt_pipelines_dir, TSourceState
from dlt.common.destination.reference import TDestinationReferenceArg
from dlt.common.schema.utils import (
    group_tables_by_resource,
    has_table_seen_data,
    is_complete_column,
    remove_defaults,
)
from dlt.common.storages import PackageStorage

from dlt.extract.state import resource_state
from dlt.pipeline.helpers import pipeline_drop
from dlt.pipeline.exceptions import CannotRestorePipelineException
from dlt._workspace.cli import echo as fmt, utils
from dlt._workspace.cli.exceptions import CliCommandException, CliCommandInnerException
from dlt._workspace.cli._urls import DLT_PIPELINE_COMMAND_DOCS_URL  # noqa: F401


def list_pipelines(pipelines_dir: str = None, verbosity: int = 1) -> None:
    """List all pipelines in the given directory, sorted by last run time.

    Args:
        pipelines_dir: Directory containing pipeline folders. If None, uses the default
                      dlt pipelines directory.
        verbosity: Controls output detail level:
                   - 0: Only show count summary
                   - 1+: Show full list with last run times
    """
    pipelines_dir, pipelines = utils.list_local_pipelines(pipelines_dir)

    if len(pipelines) > 0:
        if verbosity == 0:
            fmt.echo(
                "%s pipelines found in %s. Use %s to see the full list."
                % (len(pipelines), fmt.bold(pipelines_dir), fmt.bold("-v"))
            )
            return
        else:
            fmt.echo("%s pipelines found in %s" % (len(pipelines), fmt.bold(pipelines_dir)))
    else:
        fmt.echo("No pipelines found in %s" % fmt.bold(pipelines_dir))
        return

    # pipelines are already sorted by timestamp (newest first) from get_local_pipelines
    for pipeline_info in pipelines:
        name = pipeline_info["name"]
        timestamp = pipeline_info["timestamp"]
        time_str = utils.date_from_timestamp_with_ago(timestamp)
        fmt.echo(
            "%s %s" % (fmt.style(name, fg="green"), fmt.style(f"(last run: {time_str})", fg="cyan"))
        )


def pipeline_command(
    operation: str,
    pipeline_name: str,
    pipelines_dir: str,
    verbosity: int,
    dataset_name: str = None,
    destination: TDestinationReferenceArg = None,
    **command_kwargs: Any,
) -> None:
    if operation == "list":
        list_pipelines(pipelines_dir)
        return

    # we may open the dashboard for a pipeline without checking if it exists
    if operation == "show":
        from dlt._workspace.helpers.dashboard.runner import run_dashboard

        run_dashboard(pipeline_name, edit=command_kwargs.get("edit"), pipelines_dir=pipelines_dir)
        return

    try:
        if verbosity > 0:
            fmt.echo("Attaching to pipeline %s" % fmt.bold(pipeline_name))
        p = dlt.attach(pipeline_name=pipeline_name, pipelines_dir=pipelines_dir)
    except CannotRestorePipelineException as e:
        if operation not in {"sync", "drop"}:
            raise
        fmt.warning(str(e))
        if not fmt.confirm(
            "Do you want to attempt to restore the pipeline state from destination?",
            default=False,
        ):
            return
        destination = destination or fmt.text_input(
            f"Enter destination name for pipeline {fmt.bold(pipeline_name)}"
        )
        dataset_name = dataset_name or fmt.text_input(
            f"Enter dataset name for pipeline {fmt.bold(pipeline_name)}"
        )
        p = dlt.pipeline(
            pipeline_name,
            pipelines_dir,
            destination=destination,
            dataset_name=dataset_name,
        )
        p.sync_destination()
        if p.first_run:
            # remote state was not found
            p._wipe_working_folder()
            fmt.error(
                f"Pipeline {pipeline_name} was not found in dataset {dataset_name} in {destination}"
            )
            return
        if operation == "sync":
            return  # No need to sync again

    def _display_pending_packages() -> Tuple[Sequence[str], Sequence[str]]:
        extracted_packages = p.list_extracted_load_packages()
        if extracted_packages:
            fmt.echo(
                "Has %s extracted packages ready to be normalized with following load ids:"
                % fmt.bold(str(len(extracted_packages)))
            )
            for load_id in extracted_packages:
                fmt.echo(load_id)
        norm_packages = p.list_normalized_load_packages()
        if norm_packages:
            fmt.echo(
                "Has %s normalized packages ready to be loaded with following load ids:"
                % fmt.bold(str(len(norm_packages)))
            )
            for load_id in norm_packages:
                fmt.echo(load_id)
            # load first (oldest) package
            first_package_info = p.get_load_package_info(norm_packages[0])
            if PackageStorage.is_package_partially_loaded(first_package_info):
                fmt.warning(
                    "This package is partially loaded. Data in the destination may be modified."
                )
            fmt.echo()
        return extracted_packages, norm_packages

    # launch mcp server before outputting to stdout
    if operation == "mcp":
        from dlt._workspace.mcp import PipelineMCP

        if command_kwargs["stdio"]:
            transport = "stdio"
        elif command_kwargs.get("sse"):
            transport = "sse"
        else:
            transport = "streamable-http"
        if transport != "stdio":
            fmt.echo("Starting dlt MCP server", err=True)
        mcp = PipelineMCP(p.pipeline_name, command_kwargs["port"])
        mcp.run(transport=transport)

        return

    fmt.echo("Found pipeline %s in %s" % (fmt.bold(p.pipeline_name), fmt.bold(p.pipelines_dir)))

    if operation == "info":
        state: TSourceState = p.state  # type: ignore
        fmt.echo("Synchronized state:")
        for k, v in state.items():
            if not isinstance(v, dict):
                fmt.echo("%s: %s" % (fmt.style(k, fg="green"), v))
        sources_state = state.get("sources")
        if sources_state:
            fmt.echo()
            fmt.secho("sources:", fg="green")
            if verbosity > 0:
                fmt.echo(json.dumps(sources_state, pretty=True))
            else:
                fmt.echo("Add -v option to see sources state. Note that it could be large.")

        fmt.echo()
        fmt.echo("Local state:")
        for k, v in state["_local"].items():
            if isinstance(v, dict):
                # show run context id
                if k == "last_run_context":
                    k = "last_run_context['uri']"
                    v = v["uri"]
                else:
                    v = None
            if v is not None:
                fmt.echo("%s: %s" % (fmt.style(k, fg="green"), v))
        fmt.echo()
        if p.default_schema_name is None:
            fmt.warning("This pipeline does not have a default schema")
        else:
            is_single_schema = len(p.schema_names) == 1
            for schema_name in p.schema_names:
                fmt.echo("Resources in schema: %s" % fmt.bold(schema_name))
                schema = p.schemas[schema_name]
                data_tables = {t["name"]: t for t in schema.data_tables()}
                for resource_name, tables in group_tables_by_resource(data_tables).items():
                    res_state_slots = 0
                    if sources_state:
                        source_state = (
                            next(iter(sources_state.items()))[1]
                            if is_single_schema
                            else sources_state.get(schema_name)
                        )
                        if source_state:
                            resource_state_ = resource_state(resource_name, source_state)
                            res_state_slots = len(resource_state_)
                    fmt.echo(
                        "%s with %s table(s) and %s resource state slot(s)"
                        % (
                            fmt.bold(resource_name),
                            fmt.bold(str(len(tables))),
                            fmt.bold(str(res_state_slots)),
                        )
                    )
                    if verbosity > 0:
                        for table in tables:
                            incomplete_columns = len(
                                [
                                    col
                                    for col in table["columns"].values()
                                    if not is_complete_column(col)
                                ]
                            )
                            fmt.echo(
                                "\t%s table %s column(s) %s %s"
                                % (
                                    fmt.bold(table["name"]),
                                    fmt.bold(str(len(table["columns"]))),
                                    (
                                        fmt.style("received data", fg="green")
                                        if has_table_seen_data(table)
                                        else fmt.style("not yet received data", fg="yellow")
                                    ),
                                    (
                                        fmt.style(
                                            f"{incomplete_columns} incomplete column(s)",
                                            fg="yellow",
                                        )
                                        if incomplete_columns > 0
                                        else ""
                                    ),
                                )
                            )
        fmt.echo()
        fmt.echo("Working dir content:")
        _display_pending_packages()
        loaded_packages = p.list_completed_load_packages()
        if loaded_packages:
            fmt.echo(
                "Has %s completed load packages with following load ids:"
                % fmt.bold(str(len(loaded_packages)))
            )
            for load_id in loaded_packages:
                fmt.echo(load_id)
            fmt.echo()
        trace = p.last_trace
        if trace is None or len(trace.steps) == 0:
            fmt.echo("Pipeline does not have last run trace.")
        else:
            fmt.echo(
                "Pipeline has last run trace. Use 'dlt pipeline %s trace' to inspect "
                % pipeline_name
            )

    if operation == "trace":
        trace = p.last_trace
        if trace is None or len(trace.steps) == 0:
            fmt.warning("Pipeline does not have last run trace.")
            return
        fmt.echo(trace.asstr(verbosity))

    if operation == "failed-jobs":
        completed_loads = p.list_completed_load_packages()
        normalized_loads = p.list_normalized_load_packages()
        for load_id in completed_loads + normalized_loads:  # type: ignore
            fmt.echo("Checking failed jobs in load id '%s'" % fmt.bold(load_id))
            failed_jobs = p.list_failed_jobs_in_package(load_id)
            if failed_jobs:
                for failed_job in p.list_failed_jobs_in_package(load_id):
                    fmt.echo(
                        "JOB: %s(%s)"
                        % (
                            fmt.bold(failed_job.job_file_info.job_id()),
                            fmt.bold(failed_job.job_file_info.table_name),
                        )
                    )
                    fmt.echo("JOB file type: %s" % fmt.bold(failed_job.job_file_info.file_format))
                    fmt.echo("JOB file path: %s" % fmt.bold(failed_job.file_path))
                    if verbosity > 0:
                        fmt.echo(failed_job.asstr(verbosity))
                    fmt.secho(failed_job.failed_message, fg="red")
                    fmt.echo()
            else:
                fmt.echo("No failed jobs found")

    if operation == "drop-pending-packages":
        extracted_packages, norm_packages = _display_pending_packages()
        if len(extracted_packages) == 0 and len(norm_packages) == 0:
            fmt.echo("No pending packages found")
        if fmt.confirm("Delete the above packages?", default=False):
            p.drop_pending_packages(with_partial_loads=True)
            fmt.echo("Pending packages deleted")

    if operation == "sync":
        if fmt.confirm(
            "About to drop the local state of the pipeline and reset all the schemas. The"
            " destination state, data and schemas are left intact. Proceed?",
            default=False,
        ):
            fmt.echo("Dropping local state")
            p = p.drop()
            fmt.echo("Restoring from destination")
            p.sync_destination()
            if p.first_run:
                # remote state was not found
                p._wipe_working_folder()
                fmt.error(
                    f"Pipeline {pipeline_name} was not found in dataset {dataset_name} in"
                    f" {destination}"
                )
                return

    if operation == "load-package":
        load_id = command_kwargs.get("load_id")
        if not load_id:
            packages = sorted(p.list_extracted_load_packages())
            if not packages:
                packages = sorted(p.list_normalized_load_packages())
            if not packages:
                packages = sorted(p.list_completed_load_packages())
            if not packages:
                raise CliCommandInnerException(
                    "pipeline", "There are no load packages for that pipeline"
                )
            load_id = packages[-1]

        package_info = p.get_load_package_info(load_id)
        fmt.echo(
            "Package %s found in %s" % (fmt.bold(load_id), fmt.bold(package_info.package_path))
        )
        fmt.echo(package_info.asstr(verbosity))
        if len(package_info.schema_update) > 0:
            if verbosity == 0:
                fmt.echo("Add -v option to see schema update. Note that it could be large.")
            else:
                tables = remove_defaults({"tables": package_info.schema_update})
                fmt.echo(fmt.bold("Schema update:"))
                fmt.echo(
                    yaml.dump(
                        tables,
                        allow_unicode=True,
                        default_flow_style=False,
                        sort_keys=False,
                    )
                )

    if operation == "schema":
        if not p.default_schema_name:
            fmt.warning("Pipeline does not have a default schema")
        else:
            fmt.echo("Found schema with name %s" % fmt.bold(p.default_schema_name))
        format_ = command_kwargs.get("format")
        remove_defaults_ = command_kwargs.get("remove_defaults")
        s = p.default_schema
        export = utils.fetch_schema_export(s, format_=format_, remove_defaults=remove_defaults_)
        fmt.echo(export["content"])

    if operation == "drop":
        drop = pipeline_drop(
            p,
            resources=command_kwargs.get("resources", ()),
            schema_name=command_kwargs.get("schema_name"),
            state_paths=command_kwargs.get("state_paths", ()),
            drop_all=command_kwargs.get("drop_all", False),
            state_only=command_kwargs.get("state_only", False),
        )
        if drop.is_empty:
            fmt.echo(
                "Could not select any resources to drop and no resource/source state to reset. Use"
                " the command below to inspect the pipeline:"
            )
            fmt.echo(f"dlt pipeline -v {p.pipeline_name} info")
            if len(drop.info["warnings"]):
                fmt.echo("Additional warnings are available")
                for warning in drop.info["warnings"]:
                    fmt.warning(warning)
            return

        # drop command will fail if first run but that happens later, so we make sure that last_run_context exists
        if not p.first_run:
            from dlt.common.runtime import run_context

            active_run_dir = os.path.abspath(run_context.active().run_dir)

            if p.last_run_context["run_dir"] != active_run_dir:
                fmt.warning(
                    fmt.style(
                        "You should run this from the same directory as the pipeline script (%s),"
                        " where the folder with credentials (%s) is located. Alternatively, you can"
                        " set the required credentials as environment variables."
                        % (
                            p.last_run_context["run_dir"],
                            p.last_run_context["settings_dir"],
                        ),
                        fg="yellow",
                    )
                )

        fmt.echo(
            "About to drop the following data in dataset %s in destination %s:"
            % (
                fmt.bold(p.dataset_name),
                fmt.bold(p.destination.destination_name),
            )
        )
        fmt.echo("%s: %s" % (fmt.style("Selected schema", fg="green"), drop.info["schema_name"]))
        fmt.echo(
            "%s: %s"
            % (
                fmt.style("Selected resource(s)", fg="green"),
                drop.info["resource_names"],
            )
        )
        fmt.echo("%s: %s" % (fmt.style("Table(s) to drop", fg="green"), drop.info["tables"]))
        fmt.echo(
            "%s: %s"
            % (fmt.style("\twith data in destination", fg="green"), drop.info["tables_with_data"])
        )
        fmt.echo(
            "%s: %s"
            % (
                fmt.style("Resource(s) state to reset", fg="green"),
                drop.info["resource_states"],
            )
        )
        fmt.echo(
            "%s: %s"
            % (
                fmt.style("Source state path(s) to reset", fg="green"),
                drop.info["state_paths"],
            )
        )
        for warning in drop.info["warnings"]:
            fmt.warning(warning)
        if fmt.confirm("Do you want to apply these changes?", default=False):
            drop()


@utils.track_command("pipeline", True, "operation")
def pipeline_command_wrapper(
    operation: str, pipeline_name: str, pipelines_dir: str, verbosity: int, **command_kwargs: Any
) -> None:
    try:
        pipeline_command(operation, pipeline_name, pipelines_dir, verbosity, **command_kwargs)
    except CannotRestorePipelineException as ex:
        fmt.secho(str(ex), err=True, fg="red")
        fmt.secho(
            "Try command %s to restore the pipeline state from destination"
            % fmt.bold(f"dlt pipeline {pipeline_name} sync")
        )
        raise CliCommandException(error_code=-2)



# flake8: noqa
import os
from typing import Iterator

import dlt
from dlt.sources import TDataItems
from dlt.sources.filesystem import FileItemDict, filesystem, readers, read_csv


# where the test files are, those examples work with (url)
TESTS_BUCKET_URL = "samples"


def stream_and_merge_csv() -> None:
    """Demonstrates how to scan folder with csv files, load them in chunk and merge on date column with the previous load"""
    pipeline = dlt.pipeline(
        pipeline_name="standard_filesystem_csv",
        destination="duckdb",
        dataset_name="met_data",
    )
    # met_data contains 3 columns, where "date" column contain a date on which we want to merge
    # load all csvs in A801
    met_files = readers(bucket_url=TESTS_BUCKET_URL, file_glob="met_csv/A801/*.csv").read_csv()
    # tell dlt to merge on date
    met_files.apply_hints(write_disposition="merge", merge_key="date")
    # NOTE: we load to met_csv table
    load_info = pipeline.run(met_files.with_name("met_csv"))
    print(load_info)
    print(pipeline.last_trace.last_normalize_info)

    # now let's simulate loading on next day. not only current data appears but also updated record for the previous day are present
    # all the records for previous day will be replaced with new records
    met_files = readers(bucket_url=TESTS_BUCKET_URL, file_glob="met_csv/A801/*.csv").read_csv()
    met_files.apply_hints(write_disposition="merge", merge_key="date")
    load_info = pipeline.run(met_files.with_name("met_csv"))

    # you can also do dlt pipeline standard_filesystem_csv show to confirm that all A801 were replaced with A803 records for overlapping day
    print(load_info)
    print(pipeline.last_trace.last_normalize_info)


def read_csv_with_duckdb() -> None:
    pipeline = dlt.pipeline(
        pipeline_name="standard_filesystem",
        destination="duckdb",
        dataset_name="met_data_duckdb",
    )

    # load all the CSV data, excluding headers
    met_files = readers(
        bucket_url=TESTS_BUCKET_URL, file_glob="met_csv/A801/*.csv"
    ).read_csv_duckdb(chunk_size=1000, header=True)

    load_info = pipeline.run(met_files)

    print(load_info)
    print(pipeline.last_trace.last_normalize_info)


def read_csv_duckdb_compressed() -> None:
    pipeline = dlt.pipeline(
        pipeline_name="standard_filesystem",
        destination="duckdb",
        dataset_name="taxi_data",
        dev_mode=True,
    )

    met_files = readers(
        bucket_url=TESTS_BUCKET_URL,
        file_glob="gzip/*",
    ).read_csv_duckdb()

    load_info = pipeline.run(met_files)
    print(load_info)
    print(pipeline.last_trace.last_normalize_info)


def read_parquet_and_jsonl_chunked() -> None:
    pipeline = dlt.pipeline(
        pipeline_name="standard_filesystem",
        destination="duckdb",
        dataset_name="teams_data",
    )
    # When using the readers resource, you can specify a filter to select only the files you
    # want to load including a glob pattern. If you use a recursive glob pattern, the filenames
    # will include the path to the file inside the bucket_url.

    # JSONL reading (in large chunks!)
    jsonl_reader = readers(TESTS_BUCKET_URL, file_glob="**/*.jsonl").read_jsonl(chunksize=10000)
    # PARQUET reading
    parquet_reader = readers(TESTS_BUCKET_URL, file_glob="**/*.parquet").read_parquet()
    # load both folders together to specified tables
    load_info = pipeline.run(
        [
            jsonl_reader.with_name("jsonl_team_data"),
            parquet_reader.with_name("parquet_team_data"),
        ]
    )
    print(load_info)
    print(pipeline.last_trace.last_normalize_info)


def read_custom_file_type_excel() -> None:
    """Here we create an extract pipeline using filesystem resource and read_csv transformer"""

    # instantiate filesystem directly to get list of files (FileItems) and then use read_excel transformer to get
    # content of excel via pandas

    @dlt.transformer
    def read_excel(items: Iterator[FileItemDict], sheet_name: str) -> Iterator[TDataItems]:
        import pandas as pd

        for file_obj in items:
            with file_obj.open() as file:
                yield pd.read_excel(file, sheet_name).to_dict(orient="records")

    freshman_xls = filesystem(
        bucket_url=TESTS_BUCKET_URL, file_glob="../custom/freshman_kgs.xlsx"
    ) | read_excel("freshman_table")

    load_info = dlt.run(
        freshman_xls.with_name("freshman"),
        destination="duckdb",
        dataset_name="freshman_data",
    )
    print(load_info)


def copy_files_resource(local_folder: str) -> None:
    """Demonstrates how to copy files locally by adding a step to filesystem resource and the to load the download listing to db"""
    pipeline = dlt.pipeline(
        pipeline_name="standard_filesystem_copy",
        destination="duckdb",
        dataset_name="standard_filesystem_data",
    )

    # a step that copies files into test storage
    def _copy(item: FileItemDict) -> FileItemDict:
        # instantiate fsspec and copy file
        dest_file = os.path.join(local_folder, item["relative_path"])
        # create dest folder
        os.makedirs(os.path.dirname(dest_file), exist_ok=True)
        # download file
        item.fsspec.download(item["file_url"], dest_file)
        # return file item unchanged
        return item

    # use recursive glob pattern and add file copy step
    downloader = filesystem(TESTS_BUCKET_URL, file_glob="**").add_map(_copy)

    # NOTE: you do not need to load any data to execute extract, below we obtain
    # a list of files in a bucket and also copy them locally
    # listing = list(downloader)
    # print(listing)

    # download to table "listing"
    # downloader = filesystem(TESTS_BUCKET_URL, file_glob="**").add_map(_copy)
    load_info = pipeline.run(downloader.with_name("listing"), write_disposition="replace")
    # pretty print the information on data that was loaded
    print(load_info)
    print(pipeline.last_trace.last_normalize_info)


def read_files_incrementally_mtime() -> None:
    pipeline = dlt.pipeline(
        pipeline_name="standard_filesystem_incremental",
        destination="duckdb",
        dataset_name="file_tracker",
    )

    # here we modify filesystem resource so it will track only new csv files
    # such resource may be then combined with transformer doing further processing
    new_files = filesystem(bucket_url=TESTS_BUCKET_URL, file_glob="csv/*")
    # add incremental on modification time
    new_files.apply_hints(incremental=dlt.sources.incremental("modification_date"))
    load_info = pipeline.run((new_files | read_csv()).with_name("csv_files"))
    print(load_info)
    print(pipeline.last_trace.last_normalize_info)

    # load again - no new files!
    new_files = filesystem(bucket_url=TESTS_BUCKET_URL, file_glob="csv/*")
    # add incremental on modification time
    new_files.apply_hints(incremental=dlt.sources.incremental("modification_date"))
    load_info = pipeline.run((new_files | read_csv()).with_name("csv_files"))
    print(load_info)
    print(pipeline.last_trace.last_normalize_info)


if __name__ == "__main__":
    copy_files_resource("_storage")
    stream_and_merge_csv()
    read_parquet_and_jsonl_chunked()
    read_custom_file_type_excel()
    read_files_incrementally_mtime()
    read_csv_with_duckdb()
    read_csv_duckdb_compressed()



"""Module registering command line plugins
To add a new plugin here, do the following:
1. create a new command class in like ie. `dlt._workspace.cli.commands `ProfileCommand(SupportsCliCommand):`
2. provide the implementation of command functions like ie. in `dlt._workspace.cli._profile_command`
3. remember to wrap command in telemetry ie. @utils.track_command("profile", track_before=False, operation="info")
4. register the plugin here.

this module is inspected by pluggy on dlt startup
"""
from typing import Type

from dlt.common.configuration import plugins
from dlt.common.runtime.run_context import active as run_context_active


__all__ = [
    "plug_cli_init",
    "plug_cli_pipeline",
    "plug_cli_schema",
    "plug_cli_dashboard",
    "plug_cli_telemetry",
    "plug_cli_deploy",
    "plug_cli_docs",
    "plug_cli_ai",
    "plug_cli_profile",
    "plug_cli_workspace",
]


def is_workspace_active() -> bool:
    # verify run context type without importing

    ctx = run_context_active()
    return ctx.__class__.__name__ == "WorkspaceRunContext"


@plugins.hookimpl(specname="plug_cli")
def plug_cli_init() -> Type[plugins.SupportsCliCommand]:
    from dlt._workspace.cli.commands import InitCommand

    return InitCommand


@plugins.hookimpl(specname="plug_cli")
def plug_cli_pipeline() -> Type[plugins.SupportsCliCommand]:
    from dlt._workspace.cli.commands import PipelineCommand

    return PipelineCommand


@plugins.hookimpl(specname="plug_cli")
def plug_cli_schema() -> Type[plugins.SupportsCliCommand]:
    from dlt._workspace.cli.commands import SchemaCommand

    return SchemaCommand


@plugins.hookimpl(specname="plug_cli")
def plug_cli_dashboard() -> Type[plugins.SupportsCliCommand]:
    from dlt._workspace.cli.commands import DashboardCommand

    return DashboardCommand


@plugins.hookimpl(specname="plug_cli")
def plug_cli_telemetry() -> Type[plugins.SupportsCliCommand]:
    from dlt._workspace.cli.commands import TelemetryCommand

    return TelemetryCommand


@plugins.hookimpl(specname="plug_cli")
def plug_cli_deploy() -> Type[plugins.SupportsCliCommand]:
    from dlt._workspace.cli.commands import DeployCommand

    return DeployCommand


@plugins.hookimpl(specname="plug_cli")
def plug_cli_docs() -> Type[plugins.SupportsCliCommand]:
    from dlt._workspace.cli.commands import CliDocsCommand

    return CliDocsCommand


@plugins.hookimpl(specname="plug_cli")
def plug_cli_ai() -> Type[plugins.SupportsCliCommand]:
    from dlt._workspace.cli.commands import AiCommand

    return AiCommand


@plugins.hookimpl(specname="plug_cli")
def plug_cli_profile() -> Type[plugins.SupportsCliCommand]:
    if is_workspace_active():
        from dlt._workspace.cli.commands import ProfileCommand

        return ProfileCommand
    else:
        return None


@plugins.hookimpl(specname="plug_cli")
def plug_cli_workspace() -> Type[plugins.SupportsCliCommand]:
    if is_workspace_active():
        from dlt._workspace.cli.commands import WorkspaceCommand

        return WorkspaceCommand
    else:
        return None



import os
from typing import Dict, Any, List, Optional, Union
from pathlib import Path
import warnings
import time

from fsspec import AbstractFileSystem
from packaging.version import Version

from dlt import version
from dlt.common import logger
from dlt.common.configuration import configspec
from dlt.common.destination.exceptions import DestinationUndefinedEntity
from dlt.common.libs.pyarrow import cast_arrow_schema_types
from dlt.common.libs.utils import load_open_tables
from dlt.common.pipeline import SupportsPipeline
from dlt.common.schema.typing import TWriteDisposition, TTableSchema
from dlt.common.schema.utils import get_first_column_name_with_prop, get_columns_names_with_prop
from dlt.common.utils import assert_min_pkg_version
from dlt.common.exceptions import MissingDependencyException
from dlt.common.storages.configuration import FileSystemCredentials, FilesystemConfiguration
from dlt.common.configuration.specs import BaseConfiguration, CredentialsConfiguration
from dlt.common.configuration.specs.mixins import WithPyicebergConfig
from dlt.common.configuration.inject import with_config


try:
    import pyiceberg
    from pyiceberg.table import Table as IcebergTable
    from pyiceberg.catalog import Catalog as IcebergCatalog
    from pyiceberg.exceptions import NoSuchTableError
    from pyiceberg.partitioning import (
        UNPARTITIONED_PARTITION_SPEC,
        PartitionSpec as IcebergPartitionSpec,
    )
    import pyarrow as pa
    from pydantic import BaseModel, ConfigDict, Field
except ModuleNotFoundError:
    raise MissingDependencyException(
        "dlt pyiceberg helpers",
        [f"{version.DLT_PKG_NAME}[pyiceberg]"],
        "Install `pyiceberg` so dlt can create Iceberg tables in the `filesystem` destination.",
    )

pyiceberg_semver = Version(pyiceberg.__version__)

if pyiceberg_semver < Version("0.10.0"):
    import pyiceberg.io.pyarrow as _pio

    _orig_get_kwargs = _pio._get_parquet_writer_kwargs

    def _patched_get_parquet_writer_kwargs(table_properties):  # type: ignore[no-untyped-def]
        """Return the original kwargs **plus** store_decimal_as_integer=True."""
        kwargs = _orig_get_kwargs(table_properties)
        kwargs.setdefault("store_decimal_as_integer", True)
        return kwargs

    _pio._get_parquet_writer_kwargs = _patched_get_parquet_writer_kwargs


def ensure_iceberg_compatible_arrow_schema(schema: pa.Schema) -> pa.Schema:
    ARROW_TO_ICEBERG_COMPATIBLE_ARROW_TYPE_MAP = {
        pa.types.is_time32: pa.time64("us"),
        pa.types.is_decimal256: pa.string(),  # pyarrow does not allow downcasting to decimal128
        pa.types.is_dictionary: lambda t_: t_.value_type,
    }
    return cast_arrow_schema_types(schema, ARROW_TO_ICEBERG_COMPATIBLE_ARROW_TYPE_MAP)


def ensure_iceberg_compatible_arrow_data(data: pa.Table) -> pa.Table:
    schema = ensure_iceberg_compatible_arrow_schema(data.schema)
    return data.cast(schema)


def write_iceberg_table(
    table: IcebergTable,
    data: pa.Table,
    write_disposition: TWriteDisposition,
) -> None:
    start_ts = time.monotonic()
    if write_disposition == "append":
        table.append(ensure_iceberg_compatible_arrow_data(data))
    elif write_disposition == "replace":
        table.overwrite(ensure_iceberg_compatible_arrow_data(data))
    logger.debug(
        f"pyiceberg: {write_disposition} arrow with {data.num_rows} rows to table {table.name()} at"
        f" location {table.location()} took {(time.monotonic() - start_ts)} seconds."
    )


def merge_iceberg_table(
    table: IcebergTable,
    data: pa.Table,
    schema: TTableSchema,
    load_table_name: str,
) -> None:
    """Merges in-memory Arrow data into on-disk Iceberg table."""
    strategy = schema["x-merge-strategy"]  # type: ignore[typeddict-item]
    if strategy == "upsert":
        # evolve schema
        with table.update_schema() as update:
            update.union_by_name(ensure_iceberg_compatible_arrow_schema(data.schema))

        if "parent" in schema:
            join_cols = [get_first_column_name_with_prop(schema, "unique")]
        else:
            join_cols = get_columns_names_with_prop(schema, "primary_key")

        # TODO: replace the batching method with transaction with pyiceberg's release after 0.9.1
        for rb in data.to_batches(max_chunksize=1_000):
            batch_tbl = pa.Table.from_batches([rb])
            batch_tbl = ensure_iceberg_compatible_arrow_data(batch_tbl)

            table.upsert(
                df=batch_tbl,
                join_cols=join_cols,
                when_matched_update_all=True,
                when_not_matched_insert_all=True,
                case_sensitive=True,
            )
    else:
        raise ValueError(
            f'Merge strategy "{strategy}" is not supported for Iceberg tables. '
            f'Table: "{load_table_name}".'
        )


def get_sql_catalog(
    catalog_name: str,
    uri: str,
    credentials: FileSystemCredentials,
    properties: Dict[str, Any] = None,
) -> IcebergCatalog:  # noqa: F821
    assert_min_pkg_version(
        pkg_name="sqlalchemy",
        version="2.0.18",
        msg=(
            "`sqlalchemy>=2.0.18` is needed for `iceberg` table format on `filesystem` destination."
        ),
    )

    from pyiceberg.catalog.sql import SqlCatalog

    return SqlCatalog(
        catalog_name,
        uri=uri,
        **_get_fileio_config(credentials),
        **(properties or {}),
    )


class CatalogNotFoundError(Exception):
    """Raised when a catalog cannot be found in the specified configuration method"""

    pass


class PyicebergCatalogConfig(BaseModel):
    model_config = ConfigDict(extra="allow")

    type: str = Field(..., description="Iceberg catalog type")  # noqa
    uri: str = Field(..., description="Iceberg catalog URI")
    warehouse: str = Field(..., description="Warehouse name")


@configspec
class IcebergConfig(BaseConfiguration):
    # Iceberg catalog configuration
    iceberg_catalog_name: str = "default"
    """Name of the Iceberg catalog to use. Corresponds to catalog name in .pyiceberg.yaml"""

    iceberg_catalog_type: Optional[str] = "sql"
    """Type of Iceberg catalog: 'sql', 'rest', 'glue', 'hive', etc."""

    iceberg_catalog_config: Optional[Dict[str, Any]] = None
    """
    Optional dictionary with complete catalog configuration.
    If provided, will be used instead of loading from .pyiceberg.yaml.
    Example for REST catalog:
        {
            'type': 'rest',
            'uri': 'https://catalog.example.com',
            'warehouse': 'my_warehouse',
            'credential': 'token',
            'scope': 'PRINCIPAL_ROLE:ALL'
        }
    Example for SQL catalog:
        {
            'type': 'sql',
            'uri': 'postgresql://user:pass@localhost/catalog'
        }

    Example for secrets.toml:
        [iceberg_catalog]
        iceberg_catalog_name = "default"
        iceberg_catalog_type = "rest"

        [iceberg_catalog.iceberg_catalog_config]
        uri = "http://localhost:8181/catalog"
        warehouse = "default"
        header.X-Iceberg-Access-Delegation = "remote-signing"
        py-io-impl = "pyiceberg.io.fsspec.FsspecFileIO"
        s3.endpoint = "https://cool-bucket.com/"
        s3.access-key-id = "cool-bucket-access-key"
        s3.secret-access-key = "cool-bucket-secret-key"
        s3.region = "cool-bucket-region"
    """


def _load_catalog_from_pyiceberg(
    catalog_name: str,
) -> IcebergCatalog:
    """Load Iceberg catalog through pyiceberg load_catalog mechanism. See https://py.iceberg.apache.org/configuration/#setting-configuration-values

    Args:
        catalog_name: Name of the catalog to load from YAML

    Returns:
        IcebergCatalog instance loaded from YAML configuration

    Raises:
        CatalogNotFoundError: If no .pyiceberg.yaml file found or catalog not in file

    Search paths (in order):
        1. PYICEBERG_HOME environment variable (pyiceberg standard)
        2. DLT run directory
        3. DLT settings directory

    Example .pyiceberg.yaml:
        catalog:
          my_catalog:
            type: rest
            uri: https://catalog.example.com
            warehouse: my_warehouse
            credential: token
    """
    from pyiceberg.catalog import load_catalog
    import dlt

    active_run_context = dlt.current.run_context()

    # Search through potential paths for Iceberg Config
    search_paths = []

    pyiceberg_home = os.environ.get("PYICEBERG_HOME")
    if pyiceberg_home:
        search_paths.append(Path(pyiceberg_home) / ".pyiceberg.yaml")

    # Add dlt-specific paths
    search_paths.extend(
        [
            Path(active_run_context.run_dir) / ".pyiceberg.yaml",
            Path(active_run_context.get_setting(".pyiceberg.yaml")),
        ]
    )

    # Search for the first existing config file and confirm 'catalog:' is present
    no_config_file_found = True
    for path in search_paths:
        if path.exists():
            logger.debug(f"Searching for catalog configuration in: {path}")
            with open(path, "r", encoding="utf-8") as f:
                contents = f.read()
                if "catalog:" in contents:
                    no_config_file_found = False
                    break

    # Check if any PYICEBERG_CATALOG_* environment variable is set
    pyiceberg_env_var = any(key.startswith("PYICEBERG_CATALOG_") for key in os.environ)

    # If no config file was found, raise error
    if no_config_file_found and not pyiceberg_env_var:
        raise CatalogNotFoundError(
            "No .pyiceberg.yaml file found. Searched in:"
            f" {', '.join(str(p) for p in search_paths)}. No PYICEBERG_CATALOG_* environment"
            " variables found."
        )

    return load_catalog(catalog_name)


def _load_catalog_from_config(
    catalog_name: str,
    config_dict: Dict[str, Any],
    credentials: Optional[FileSystemCredentials] = None,
) -> IcebergCatalog:
    """Load Iceberg catalog from configuration dictionary

    Args:
        catalog_name: Name of the catalog
        config_dict: Dictionary with catalog configuration (type, uri, warehouse, etc.)

    Returns:
        IcebergCatalog instance

    Raises:
        CatalogNotFoundError: If config_dict is None or empty

    Example:
        config = {
            'type': 'rest',
            'uri': 'https://catalog.example.com',
            'warehouse': 'my_warehouse',
            'credential': 'token'
        }
        catalog = load_catalog_from_config('my_catalog', config)
    """
    from pyiceberg.catalog import load_catalog

    # Validate config
    PyicebergCatalogConfig(**config_dict)

    if not config_dict:
        raise CatalogNotFoundError("No configuration dictionary provided")

    logger.info(f"Loading catalog '{catalog_name}' from provided configuration")

    if credentials:
        config_dict.update(_get_fileio_config(credentials))

    return load_catalog(catalog_name, **config_dict)


@with_config(spec=IcebergConfig, sections="iceberg_catalog")
def get_catalog(
    iceberg_catalog_name: str = "default",
    iceberg_catalog_type: Optional[str] = None,
    iceberg_catalog_config: Optional[Dict[str, Any]] = None,
    credentials: Optional[FileSystemCredentials] = None,
) -> IcebergCatalog:
    """Get an Iceberg catalog using multiple configuration methods.

    This function tries to load a catalog in the following priority order:
    1. From explicit config dictionary (if iceberg_catalog_config provided)
    2. From .pyiceberg.yaml file or from environment variables (PYICEBERG_*). Resolved by pyiceberg load_catalog mechanism. See https://py.iceberg.apache.org/configuration/#setting-configuration-values
    4. Fall back to in-memory SQLite catalog

    Args:
        iceberg_catalog_name: Name of the catalog (default: "default")
        iceberg_catalog_type: Type of catalog ('sql' or 'rest')
        iceberg_catalog_config: Optional dictionary with complete catalog configuration
        credentials: Optional filesystem credentials. This is ONLY used for backward compatibility with in-memory SQLite catalog.

    Returns:
        IcebergCatalog instance

    Examples:

        # Load from config dict
        config = {'type': 'rest', 'uri': 'https://...', 'warehouse': 'wh'}
        catalog = get_catalog('my_catalog', iceberg_catalog_type='rest', iceberg_catalog_config=config)

        # Load from .pyiceberg.yaml
        catalog = get_catalog('my_catalog', iceberg_catalog_type='sql')

        # Load from environment variables
        # (set PYICEBERG_CATALOG_TYPE, PYICEBERG_CATALOG_URI, etc.)
        catalog = get_catalog('my_catalog', iceberg_catalog_type='rest')

    """
    logger.info(f"Attempting to load Iceberg catalog: {iceberg_catalog_name}")

    # Validate catalog type
    supported_catalog_types = ["sql", "rest"]
    if iceberg_catalog_type not in supported_catalog_types:
        raise ValueError(f"Unsupported catalog type: {iceberg_catalog_type}. Use 'sql' or 'rest'.")

    # Priority 1: Explicit config dictionary (most specific and comes from secrets.toml)
    if iceberg_catalog_config:
        try:
            return _load_catalog_from_config(iceberg_catalog_name, iceberg_catalog_config)
        except CatalogNotFoundError as e:
            logger.warning(f"Failed to load catalog from config dict: {e}")

    # Priority 2: .pyiceberg.yaml file (PyIceberg standard)
    try:
        return _load_catalog_from_pyiceberg(iceberg_catalog_name)
    except CatalogNotFoundError as e:
        logger.debug(f"Catalog not found in .pyiceberg.yaml: {e}")

    # Priority 3: Fall back to in-memory SQLite (backward compatibility)
    logger.info(
        "No catalog configuration found, using in-memory SQLite catalog (backward compatibility)"
    )
    return get_sql_catalog(iceberg_catalog_name, "sqlite:///:memory:", credentials)


def evolve_table(
    catalog: IcebergCatalog,
    fs_client: AbstractFileSystem,
    config: FilesystemConfiguration,
    table_id: str,
    table_location: str,
    schema: Optional[pa.Schema] = None,
) -> IcebergTable:
    try:
        table = catalog.load_table(table_id)
    except NoSuchTableError:
        # add table to catalog
        metadata_path = f"{table_location.rstrip('/')}/metadata"
        if fs_client.exists(metadata_path):
            # found metadata; register existing table
            table = register_table(table_id, metadata_path, catalog, fs_client, config)
        else:
            raise

    # evolve schema
    if schema is not None:
        with table.update_schema() as update:
            update.union_by_name(ensure_iceberg_compatible_arrow_schema(schema))

    return table


def create_table(
    catalog: IcebergCatalog,
    table_id: str,
    table_location: str,
    schema: Union[pa.Schema, "pyiceberg.schema.Schema"],
    partition_columns: Optional[List[str]] = None,
    partition_spec: Optional[IcebergPartitionSpec] = UNPARTITIONED_PARTITION_SPEC,
    properties: Optional[Dict[str, str]] = None,
) -> None:
    if isinstance(schema, pa.Schema):
        schema = ensure_iceberg_compatible_arrow_schema(schema)

    if partition_columns:
        warnings.warn(
            "partition_columns is deprecated. Use partition_spec instead.", DeprecationWarning
        )
        with catalog.create_table_transaction(
            table_id,
            schema=schema,
            location=table_location,
            properties=properties or {},
        ) as txn:
            # add partitioning
            with txn.update_spec() as update_spec:
                for col in partition_columns:
                    update_spec.add_identity(col)
    else:
        catalog.create_table(
            identifier=table_id,
            schema=schema,
            location=table_location,
            partition_spec=partition_spec,
            properties=properties or {},
        )


def get_iceberg_tables(
    pipeline: SupportsPipeline,
    *tables: str,
    schema_name: Optional[str] = None,
    include_dlt_tables: bool = False,
) -> Dict[str, IcebergTable]:
    """Returns Iceberg tables in `pipeline.default_schema (default)` or `schema_name` as `pyiceberg.Table` objects.

    Returned object is a dictionary with table names as keys and `Tables` objects as values.
    Optionally filters dictionary by table names specified as `*tables*`.
    Raises ValueError if table name specified as `*tables` is not found. You may try to switch to other
    schemas via `schema_name` argument.
    """
    return load_open_tables(
        pipeline, "iceberg", *tables, schema_name=schema_name, include_dlt_tables=include_dlt_tables
    )


def _get_fileio_config(credentials: CredentialsConfiguration) -> Dict[str, Any]:
    if isinstance(credentials, WithPyicebergConfig):
        return credentials.to_pyiceberg_fileio_config()
    return {}


def get_last_metadata_file(
    metadata_path: str, fs_client: AbstractFileSystem, config: FilesystemConfiguration
) -> str:
    # TODO: read version-hint.txt first and save it in filesystem
    try:
        metadata_files = [f for f in fs_client.ls(metadata_path) if f.endswith(".json")]
    except FileNotFoundError:
        raise DestinationUndefinedEntity(FileNotFoundError(metadata_path))
    if len(metadata_files) == 0:
        raise DestinationUndefinedEntity(FileNotFoundError(metadata_path))
    return make_location(sorted(metadata_files)[-1], config)


def register_table(
    identifier: str,
    metadata_path: str,
    catalog: IcebergCatalog,
    fs_client: AbstractFileSystem,
    config: FilesystemConfiguration,
) -> IcebergTable:
    last_metadata_file = get_last_metadata_file(metadata_path, fs_client, config)
    return catalog.register_table(identifier, last_metadata_file)


def make_location(path: str, config: FilesystemConfiguration) -> str:
    # don't use file protocol for local files because duckdb does not support it
    # https://github.com/duckdb/duckdb/issues/13669
    location = config.make_url(path)
    if config.is_local_filesystem and os.name == "nt":
        # pyiceberg cannot deal with windows absolute urls
        location = location.replace("file:///", "file://")
    return location



import binascii
from os.path import isfile, join
from typing import Any, ClassVar, Optional, IO

from dlt.common.typing import TSecretStrValue
from dlt.common.utils import encoding_for_mode, reveal_pseudo_secret
from dlt.common.configuration.specs.base_configuration import BaseConfiguration, configspec
from dlt.common.configuration.exceptions import ConfigFileNotFoundException
from dlt.common.runtime.exec_info import platform_supports_threading


@configspec
class RuntimeConfiguration(BaseConfiguration):
    pipeline_name: Optional[str] = None
    """Pipeline name used as component in logging, must be explicitly set"""
    sentry_dsn: Optional[str] = None  # keep None to disable Sentry
    slack_incoming_hook: Optional[TSecretStrValue] = None
    dlthub_telemetry: bool = True  # enable or disable dlthub telemetry
    dlthub_telemetry_endpoint: Optional[str] = "https://telemetry.scalevector.ai"
    dlthub_telemetry_segment_write_key: Optional[str] = None
    log_format: str = (
        "{asctime}|[{levelname}]|{process}|{thread}|{name}|{filename}|{funcName}:{lineno}|{message}"
    )
    log_level: str = "WARNING"
    request_timeout: float = 60
    """Timeout for http requests"""
    request_max_attempts: int = 5
    """Max retry attempts for http clients"""
    request_backoff_factor: float = 1
    """Multiplier applied to exponential retry delay for http requests"""
    request_max_retry_delay: float = 300
    """Maximum delay between http request retries"""
    config_files_storage_path: str = "/run/config/"
    """Platform connection"""
    dlthub_dsn: Optional[TSecretStrValue] = None
    run_id: Optional[str] = None
    http_show_error_body: bool = False
    """Include HTTP response body in raised exceptions/logs. Default is False"""
    http_max_error_body_length: int = 8192
    """Maximum length of HTTP error response body to include in logs/exceptions"""

    __section__: ClassVar[str] = "runtime"

    def on_resolved(self) -> None:
        if self.slack_incoming_hook:
            # it may be obfuscated base64 value
            # TODO: that needs to be removed ASAP
            try:
                self.slack_incoming_hook = TSecretStrValue(
                    reveal_pseudo_secret(self.slack_incoming_hook, b"dlt-runtime-2022")
                )
            except binascii.Error:
                # just keep the original value
                pass

        # telemetry uses threading
        if not platform_supports_threading():
            self.dlthub_telemetry = False

    def has_configuration_file(self, name: str) -> bool:
        return isfile(self.get_configuration_file_path(name))

    def open_configuration_file(self, name: str, mode: str) -> IO[Any]:
        path = self.get_configuration_file_path(name)
        if not self.has_configuration_file(name):
            raise ConfigFileNotFoundException(path)
        return open(path, mode, encoding=encoding_for_mode(mode))

    def get_configuration_file_path(self, name: str) -> str:
        return join(self.config_files_storage_path, name)


# backward compatibility
RunConfiguration = RuntimeConfiguration



from dlt.destinations.impl.lancedb.configuration import LanceDBClientConfiguration
from typing import cast, Any

from dlt.common.exceptions import MissingDependencyException, ValueErrorWithKnownValues
from dlt.common.destination import TDestinationReferenceArg, Destination
from dlt.common.destination.client import JobClientBase
from dlt.common.schema import Schema
from dlt.common.storages.configuration import FilesystemConfiguration
from dlt.common.libs.sqlglot import TSqlGlotDialect

from dlt.destinations.impl.athena.configuration import AthenaClientConfiguration
from dlt.destinations.impl.duckdb.configuration import DuckDbClientConfiguration
from dlt.destinations.impl.databricks.configuration import DatabricksClientConfiguration
from dlt.destinations.impl.ducklake.configuration import DuckLakeClientConfiguration
from dlt.destinations.impl.motherduck.configuration import MotherDuckClientConfiguration
from dlt.destinations.impl.postgres.configuration import PostgresClientConfiguration
from dlt.destinations.impl.redshift.configuration import RedshiftClientConfiguration
from dlt.destinations.impl.snowflake.configuration import SnowflakeClientConfiguration
from dlt.destinations.impl.mssql.configuration import MsSqlClientConfiguration
from dlt.destinations.impl.bigquery.configuration import BigQueryClientConfiguration
from dlt.destinations.impl.clickhouse.configuration import ClickHouseClientConfiguration
from dlt.destinations.impl.synapse.configuration import SynapseClientConfiguration
from dlt.destinations.impl.fabric.configuration import FabricClientConfiguration

try:
    import ibis  # noqa: I251
    import sqlglot
    import sqlglot.expressions as sge
    from ibis import BaseBackend, Expr, Table  # noqa: I251
    import ibis.backends.sql.compilers as sc
    from ibis.backends.sql.compilers.base import SQLGlotCompiler
except ImportError:
    raise MissingDependencyException("dlt ibis helpers", ["ibis-framework"])


# Map dlt data types to ibis data types
DATA_TYPE_MAP = {
    "text": "string",
    "double": "float64",
    "bool": "boolean",
    "timestamp": "timestamp",
    "bigint": "int64",
    "binary": "binary",
    "json": "string",  # Store JSON as string in ibis
    "decimal": "decimal",
    "wei": "int64",  # Wei is a large integer
    "date": "date",
    "time": "time",
}


def create_ibis_backend(
    destination: TDestinationReferenceArg, client: JobClientBase, read_only: bool = False
) -> BaseBackend:
    """Create a given ibis backend for a destination client and dataset."""

    # ensure destination is a Destination instance
    if not isinstance(destination, Destination):
        destination = Destination.from_reference(destination)

    if issubclass(destination.spec, DuckDbClientConfiguration) or issubclass(
        destination.spec, MotherDuckClientConfiguration
    ):
        from dlt.destinations.impl.duckdb.duck import DuckDbClient

        assert isinstance(client, DuckDbClient)
        # do not set read_only flag on motherduck, it is managed on server side
        if not issubclass(destination.spec, MotherDuckClientConfiguration):
            client.config.credentials.read_only = read_only
        # this will open connection to duckdb, take a clone and close the clone
        with client:
            # make sure we can access tables from current dataset without qualification
            # also prevents empty duckdb files from being created
            client.sql_client.use_dataset()
            # move main connection ownership to ibis
            con = ibis.duckdb.from_connection(client.config.credentials.conn_pool.move_conn())
    elif issubclass(destination.spec, DuckLakeClientConfiguration):
        from dlt.destinations.impl.ducklake.ducklake import DuckLakeClient

        assert isinstance(client, DuckLakeClient)
        # open connection but do not close it, ducklake always creates a separate connection
        # and will not close it in destructor
        conn = client.sql_client.open_connection()
        try:
            # make sure we can access tables from current dataset without qualification
            # also prevents empty duckdb files from being created
            client.sql_client.use_dataset()
        except Exception:
            # close explicitly, won't be done by the conn pool
            client.sql_client.close_connection()
            raise
        con = ibis.duckdb.from_connection(conn)
    elif issubclass(destination.spec, PostgresClientConfiguration):
        from dlt.destinations.impl.postgres.postgres import PostgresClient
        from dlt.destinations.impl.redshift.redshift import RedshiftClient

        assert isinstance(client, (PostgresClient, RedshiftClient))
        if destination.spec is RedshiftClientConfiguration:
            # patch psycopg
            try:
                import psycopg  # type: ignore[import-not-found, unused-ignore]

                old_fetch = psycopg.types.TypeInfo.fetch

                def _ignore_hstore(conn: Any, name: Any) -> Any:
                    if name == "hstore":
                        raise TypeError("HSTORE")
                    return old_fetch(conn, name)

                psycopg.types.TypeInfo.fetch = _ignore_hstore  # type: ignore[method-assign, unused-ignore]
            except Exception:
                pass
            # check ibis version and raise an error if it's >= 0.10.4
            ibis_version = tuple(map(int, ibis.__version__.split(".")))
            if ibis_version >= (0, 10, 4):
                raise NotImplementedError(
                    "Redshift is not properly supported by ibis as of version 0.10.4. "
                    "Please use an older version of ibis."
                )
        credentials = client.config.credentials.copy()
        # schema must be passed at path, `schema` argument does not work (overridden, probably a bug)
        credentials.database = credentials.database + "/" + client.sql_client.dataset_name
        con = ibis.connect(credentials.to_native_representation())
    elif issubclass(destination.spec, SnowflakeClientConfiguration):
        from dlt.destinations.impl.snowflake.snowflake import SnowflakeClient

        assert isinstance(client, SnowflakeClient)
        sn_credentials = client.config.credentials.to_connector_params()
        dataset_name = client.sql_client.fully_qualified_dataset_name()
        con = ibis.snowflake.connect(
            schema=dataset_name, **sn_credentials, create_object_udfs=False
        )
    elif issubclass(destination.spec, MsSqlClientConfiguration) and not issubclass(
        destination.spec, (SynapseClientConfiguration, FabricClientConfiguration)
    ):  # exclude synapse and fabric
        from dlt.destinations.impl.mssql.mssql import MsSqlJobClient

        assert isinstance(client, MsSqlJobClient)
        ms_credentials = client.config.credentials.to_native_representation()
        ms_credentials = ms_credentials.replace("synapse://", "mssql://")
        con = ibis.connect(ms_credentials, driver=client.config.credentials.driver)
    elif issubclass(destination.spec, BigQueryClientConfiguration):
        from dlt.destinations.impl.bigquery.bigquery import BigQueryClient

        assert isinstance(client, BigQueryClient)
        credentials = client.config.credentials.to_native_credentials()
        con = ibis.bigquery.connect(
            credentials=credentials,
            project_id=client.sql_client.project_id,
            dataset_id=client.sql_client.fully_qualified_dataset_name(quote=False),
            location=client.sql_client.location,
        )
    elif issubclass(destination.spec, ClickHouseClientConfiguration):
        from dlt.destinations.impl.clickhouse.clickhouse import ClickHouseClient

        ch_client = cast(ClickHouseClient, client)
        con = ibis.clickhouse.connect(
            host=ch_client.config.credentials.host,
            port=ch_client.config.credentials.http_port,
            database=ch_client.config.credentials.database,
            user=ch_client.config.credentials.username,
            password=ch_client.config.credentials.password,
            secure=bool(ch_client.config.credentials.secure),
            # compression=True,
        )
    elif issubclass(destination.spec, DatabricksClientConfiguration):
        from dlt.destinations.impl.databricks.databricks import DatabricksClient

        bricks_client = cast(DatabricksClient, client)
        con = ibis.databricks.connect(
            **bricks_client.config.credentials.to_connector_params(),
            schema=bricks_client.sql_client.dataset_name,
        )
    elif issubclass(destination.spec, AthenaClientConfiguration):
        from dlt.destinations.impl.athena.athena import AthenaClient

        athena_client = cast(AthenaClient, client)
        con = ibis.athena.connect(
            schema_name=athena_client.sql_client.dataset_name,
            **athena_client.config.to_connector_params(),
        )
    # TODO: allow for sqlalchemy mysql and sqlite here
    elif issubclass(destination.spec, FilesystemConfiguration):
        from dlt.destinations.impl.filesystem.sql_client import (
            FilesystemClient,
            FilesystemSqlClient,
        )

        # we create an in memory duckdb and create the ibis backend from it
        fs_client = cast(FilesystemClient, client)
        sql_client = fs_client.sql_client
        assert isinstance(sql_client, FilesystemSqlClient)
        # do not use context manager to not return and close the cloned connection
        duckdb_conn = sql_client.open_connection()
        # make all tables available here
        # NOTE: we should probably have the option for the user to only select a subset of tables here
        sql_client.create_views_for_all_tables()
        # why this works now: whenever a clone of connection is made, all SET commands
        # apply only to it. old code was setting `curl` on the internal clone of sql_client
        # now we export this clone directly to ibis to it works
        con = ibis.duckdb.from_connection(duckdb_conn)
        # disable destructor
        fs_client.sql_client = None
        sql_client.memory_db = None
        del sql_client
    elif issubclass(destination.spec, LanceDBClientConfiguration):
        con = _create_ibis_backend_lancedb(client)
    else:
        # NOTE: Athena could theoretically work with trino backend, but according to
        # https://github.com/ibis-project/ibis/issues/7682 connecting with aws credentials
        # does not work yet.
        raise NotImplementedError(
            f"Destination type `{Destination.from_reference(destination).destination_type}` is not"
            " supported by the Ibis backend."
        )

    return con


def _create_ibis_backend_lancedb(client: JobClientBase) -> BaseBackend:
    from dlt.destinations.impl.lancedb.lancedb_client import LanceDBClient
    from dlt.destinations.impl.lancedb.sql_client import (
        get_lance_table_uri,
        _prepare_create_view_statement,
    )

    assert isinstance(client, LanceDBClient)
    # open connection but do not close it, ducklake always creates a separate connection
    # and will not close it in destructor
    for table_name in client.schema.tables:
        client.sql_client.create_view(table_name)

    native_con = client.sql_client.open_connection()
    return ibis.duckdb.from_connection(native_con)


def create_unbound_ibis_table(schema: Schema, dataset_name: str, table_name: str) -> Table:
    """Create an unbound ibis table from a dlt schema. No additional identifiers normalization, quoting
    or escaping is performed.
    """
    table_schema = schema.tables[table_name]

    # Convert dlt table schema columns to ibis schema
    ibis_schema = {
        col_name: DATA_TYPE_MAP[col_info.get("data_type", "text")]
        for col_name, col_info in table_schema.get("columns", {}).items()
    }

    # create unbound ibis table and return in dlt wrapper
    unbound_table = ibis.table(schema=ibis_schema, name=table_name, database=dataset_name)

    return unbound_table


def _get_ibis_to_sqlglot_compiler(dialect: TSqlGlotDialect) -> SQLGlotCompiler:
    """Get the compiler for a given dialect."""
    if dialect == "athena":
        compiler = sc.AthenaCompiler()
    elif dialect == "bigquery":
        compiler = sc.BigQueryCompiler()
    elif dialect == "clickhouse":
        compiler = sc.ClickHouseCompiler()
    elif dialect == "databricks":
        compiler = sc.DatabricksCompiler()
    elif dialect == "druid":
        compiler = sc.DruidCompiler()
    elif dialect == "duckdb":
        compiler = sc.DuckDBCompiler()
    elif dialect == "fabric":
        compiler = sc.MSSQLCompiler()
    elif dialect == "mysql":
        compiler = sc.MySQLCompiler()
    elif dialect == "oracle":
        compiler = sc.OracleCompiler()
    elif dialect == "postgres":
        compiler = sc.PostgresCompiler()
    elif dialect == "presto":
        compiler = sc.TrinoCompiler()
    elif dialect == "redshift":
        compiler = sc.PostgresCompiler()
    elif dialect == "risingwave":
        compiler = sc.RisingWaveCompiler()
    elif dialect == "snowflake":
        compiler = sc.SnowflakeCompiler()
    # NOTE I'm unsure if both `spark` and `spark2` are supported by the same compiler
    elif dialect == "spark":
        compiler = sc.PySparkCompiler()
    elif dialect == "spark2":
        compiler = sc.PySparkCompiler()
    elif dialect == "sqlite":
        compiler = sc.SQLiteCompiler()
    elif dialect == "trino":
        compiler = sc.TrinoCompiler()
    elif dialect == "tsql":
        compiler = sc.MSSQLCompiler()
    else:
        compiler = sc.DuckDBCompiler()

    return compiler


def compile_ibis_to_sqlglot(ibis_expr: Expr, dialect: TSqlGlotDialect) -> sge.Query:
    """Compile an ibis expression to a sqlglot query."""
    compiler = _get_ibis_to_sqlglot_compiler(dialect)
    return cast(sge.Query, compiler.to_sqlglot(ibis_expr))



"""Build a graphviz graph representation using raw strings to avoid dependency on `graphviz`"""
from __future__ import annotations

import pathlib
import textwrap
from typing import TYPE_CHECKING, Any, Optional, Union

import dlt
from dlt.destinations.dataset import ReadableDBAPIDataset
from dlt.common.exceptions import MissingDependencyException, TypeErrorWithKnownTypes
from dlt.common.schema.typing import (
    C_DLT_LOAD_ID,
    LOADS_TABLE_NAME,
    VERSION_TABLE_NAME,
    TColumnSchema,
    TSchemaTables,
    TStoredSchema,
    TTableSchema,
    TTableReference,
)
from dlt.common.schema.utils import (
    create_load_table_reference,
    create_parent_child_reference,
    create_root_child_reference,
    create_version_and_loads_hash_reference,
    create_version_and_loads_schema_name_reference,
    get_data_and_dlt_tables,
    get_first_column_name_with_prop,
    group_tables_by_resource,
    is_complete_column,
    is_nested_table,
)

if TYPE_CHECKING:
    import graphviz  # type: ignore[import-untyped]


__all__ = (
    "schema_to_graphviz",
    "render_schema_with_graphviz",
)


# TODO create a TStylesheet specs: alpha background, hints to include
INDENT = "    "
TABLE_HEADER_PORT = "p0"
TABLE_HEADER_COLOR = "#bbca06"
TABLE_HEADER_FONT_COLOR = "#1c1c34"
TABLE_BORDER_COLOR = "#1c1c34"
TABLE_ROW_COLOR = "#e7e2dd"


def _to_column_html(column: TColumnSchema, column_idx: int) -> str:
    """Generate a segment of the node label using Graphviz HTML-like language

    ref: https://graphviz.org/doc/info/shapes.html#html
    """
    column_name = column.get("name")
    if column.get("primary_key") is True:
        column_name = f"<B>{column_name}🔑</B>"

    column_type = f'{column.get("data_type")}'
    if column.get("nullable") is False:
        column_type += " <B>NN</B>"

    return f"""<tr>
            <td align="left" port="f{column_idx}" bgcolor="{TABLE_ROW_COLOR}">
                <table cellpadding="0" cellspacing="0" border="0">
                    <tr>
                        <td align="left">{column_name}</td>
                        <td align="right"><font>{column_type}</font></td>
                    </tr>
                </table>
            </td>
        </tr>"""


def _get_table_header(table_name: str) -> str:
    return f"""\
        <tr>
            <td port="{TABLE_HEADER_PORT}" bgcolor="{TABLE_HEADER_COLOR}">
                <font color="{TABLE_HEADER_FONT_COLOR}"><b>{table_name}</b></font>
            </td>
        </tr>
"""


def _to_table_html(table: TTableSchema) -> str:
    """Generate node label using Graphviz HTML-like language

    ref: https://graphviz.org/doc/info/shapes.html#html
    """
    table_header = _get_table_header(table["name"])
    dot_columns = ""
    for idx, column in enumerate(
        c for c in table.get("columns", {}).values() if is_complete_column(c)
    ):
        dot_columns += _to_column_html(column, idx + 1)

    # table header followed by columns
    return f"""\
    <table border="0" color="{TABLE_BORDER_COLOR}" cellborder="1" cellspacing="0" cellpadding="6">
        {table_header}
        {dot_columns}
    </table>
"""


def _to_dot_table(table: TTableSchema) -> str:
    """Generate graph nodes representing tables and columns"""
    name = table.get("name")
    description = f'tooltip="{table.get("description")}";' if table.get("description") else " "
    label = _to_table_html(table)
    return f'"{name}" [id="{name}";{description}label=<\n{label}>];\n\n'


def _to_dot_reference(
    *,
    from_table_name: str,
    reference: TTableReference,
    tables: TSchemaTables,
    cardinality: str = "both",
) -> str:
    from_table = tables.get(from_table_name)
    from_table_port = f"{from_table_name}:{TABLE_HEADER_PORT}"
    from_column_name = reference["columns"][0]
    # indexing starts at 1 because 0 is table header
    from_column_idx = 1 + list(from_table["columns"].keys()).index(from_column_name)
    from_column_port = f"{from_table_name}:f{from_column_idx}"

    to_table_name = reference["referenced_table"]
    to_table = tables.get(to_table_name)
    to_table_port = f"{to_table_name}:{TABLE_HEADER_PORT}"
    to_column_name = reference["referenced_columns"][0]
    # indexing starts at 1 because 0 is table header
    to_column_idx = 1 + list(to_table["columns"].keys()).index(to_column_name)
    to_column_port = f"{to_table_name}:f{to_column_idx}"

    # TODO properly handle cardinality via arrowtail and arrowhead
    # NOTE can change layout by specifying port position: n, ne, e, se, s, sw, w, nw, w, c, _
    return textwrap.dedent(
        # we need invisible edges between tables with large weight to guide the layout
        f"""{INDENT}{from_table_port} -> {to_table_port} [style=invis]
    {from_column_port}:_ -> {to_column_port}:_ [dir={cardinality}, penwidth=1, color="{TABLE_BORDER_COLOR}", arrowtail="vee", arrowhead="dot"];
    """
    )


# NOTE if changing `rankdir` to `TB`, need to change how edges are built
def _get_graph_header(schema_name: str, include_dlt_tables: bool) -> str:
    # don't forget the double `{{` to escape `{` inside an f-string; your IDE might wrongly complain
    """
    ref for graph layout: https://graphviz.org/pdf/dot.1.pdf
    Good options:
    - `layout="twopi", ranksep=5, root="_dlt_loads"`
        `ranksep` should be increased to accomodate large tables
        `root` can be changed to a more central node
    - `layout=circo` is efficient, but will have edges overlapping
    - `layout=fdp` will minimize overlap while reducing sprawl, but odd layouts

    Another important property is how stable the layout is when modifying the
    graph.
    """
    layout_root_table = f', root="{LOADS_TABLE_NAME}"' if include_dlt_tables else ""
    return f"""digraph {schema_name} {{
    graph [fontname="helvetica", fontcolor="{{TABLE_BORDER_COLOR}}", rankdir="BT", ranksep=5, layout="twopi"{layout_root_table}];
    node [penwidth=0, margin=0, fontname="helvetica"];
    edge [fontname="helvetica", fontcolor="{{TABLE_BORDER_COLOR}}", color="{{TABLE_BORDER_COLOR}}"];

"""


def _add_tables(schema: TStoredSchema, graphviz_dot: str, *, include_dlt_tables: bool) -> str:
    """Append a DOT table for each dlt table and return the updated DOT string.

    The DOT string will be in an invalid state until the graph definition is closed by `}`
    """
    data_tables, dlt_tables = get_data_and_dlt_tables(schema["tables"])

    if include_dlt_tables is True:
        # order allows to keep _dlt tables at the end
        tables = data_tables + dlt_tables
    else:
        tables = data_tables

    for table in tables:
        # skip tables with no complete columns
        if not any(is_complete_column(c) for c in table.get("columns", {}).values()):
            continue

        dot_table = _to_dot_table(table)
        graphviz_dot += dot_table

    return graphviz_dot


def _get_cluster_header(cluster_idx: int, resource_name: str) -> str:
    # don't forget the double `{{` to escape `{` inside an f-string; your IDE might wrongly complain
    # to create a cluster, the subgraph name
    # must have the `cluster_` prefix
    return f"""subgraph cluster_{cluster_idx} {{
    label="{resource_name}"

"""


def _add_table_clusters(
    schema: TStoredSchema,
    graphviz_dot: str,
    *,
    include_dlt_tables: bool,
) -> str:
    _, dlt_tables = get_data_and_dlt_tables(schema["tables"])

    cluster_idx = 0
    for cluster_idx, (resource, tables) in enumerate(
        group_tables_by_resource(schema["tables"]).items()
    ):
        if resource in ["_dlt_loads", "_dlt_version", "_dlt_pipeline_state"]:
            continue

        graphviz_dot += _get_cluster_header(cluster_idx, resource_name=resource)
        for table in tables:
            if not any(is_complete_column(c) for c in table.get("columns", {}).values()):
                continue
            graphviz_dot += INDENT + _to_dot_table(table)

        graphviz_dot += "}"

    if include_dlt_tables:
        graphviz_dot += _get_cluster_header(cluster_idx + 1, resource_name="_dlt")
        for table in dlt_tables:
            if not any(is_complete_column(c) for c in table.get("columns", {}).values()):
                continue

            graphviz_dot += INDENT + _to_dot_table(table)

        graphviz_dot += "}"

    return graphviz_dot


def _add_references(
    schema: TStoredSchema,
    graphviz_dot: str,
    *,
    include_dlt_tables: bool,
    include_internal_dlt_ref: bool,
    include_parent_child_ref: bool,
    include_root_child_ref: bool,
) -> str:
    """Append a DOT reference for each dlt table and return the updated DOT string.

    The DOT string will be in an invalid state until the graph definition is closed by `}`
    """
    tables: TSchemaTables = schema["tables"]
    for table in tables.values():
        table_name: str = table["name"]
        for reference in table.get("references", []):
            # user-defined references can have arbitrary cardinality and may incorrectly describe the data
            graphviz_dot += _to_dot_reference(
                from_table_name=table_name,
                reference=reference,
                tables=tables,
                cardinality="both",  # because it's the loosest
            )

        # link root -> loads table
        if (
            include_dlt_tables
            and include_internal_dlt_ref
            and bool(table["columns"].get(C_DLT_LOAD_ID))
        ):
            # root table contains 1 to many rows associated with a single row in loads table
            # possible cardinality: `-` (1-to-1) or `>` (m-to-1)
            graphviz_dot += _to_dot_reference(
                from_table_name=table_name,
                reference=create_load_table_reference(table),
                tables=tables,
                # cardinality=">",  # m-to-1
            )

        # link child -> parent
        if include_dlt_tables and include_parent_child_ref and is_nested_table(table):
            # child table contains 1 to many rows associated with has a single row in parent table
            # possible cardinality: `-` (1-to-1) or `>` (m-to-1)
            graphviz_dot += _to_dot_reference(
                from_table_name=table_name,
                reference=create_parent_child_reference(tables, table_name),
                tables=tables,
                # cardinality=">",  # m-to-1
            )

        # link child -> root
        if (
            include_root_child_ref
            and is_nested_table(table)
            # the table must have a root key column; can be enabled via `@dlt.source(root_key=True)` or write_disposition
            and get_first_column_name_with_prop(table, "root_key")
        ):
            # child table contains 1 to many rows associated with has a single row in root table
            # possible cardinality: `-` (1-to-1) or `>` (m-to-1)
            graphviz_dot += _to_dot_reference(
                from_table_name=table_name,
                reference=create_root_child_reference(tables, table_name),
                tables=tables,
            )

    # generate links between internal dlt tables
    if include_dlt_tables and include_internal_dlt_ref:
        # a schema version hash can have multiple runs in the loads table
        # schema version hash is unique
        # possible: cardinality: `-` (1-to-1) or `<` (1-to-m)
        graphviz_dot += _to_dot_reference(
            from_table_name=VERSION_TABLE_NAME,
            reference=create_version_and_loads_hash_reference(tables),
            tables=tables,
            # cardinality="<",
        )
        # a schema name can have multiple multiple runs in the loads table
        # schema name is not unique; it can have multiple version hash
        # possible: cardinality: `-` (1-to-1), `<` (1-to-m), or `<>` (m-to-n)
        graphviz_dot += _to_dot_reference(
            from_table_name=VERSION_TABLE_NAME,
            reference=create_version_and_loads_schema_name_reference(tables),
            tables=tables,
            # cardinality="<>",
        )

    return graphviz_dot


# most of this module reuses logic from dbml renderer
def schema_to_graphviz(
    schema: TStoredSchema,
    *,
    include_dlt_tables: bool = True,
    include_internal_dlt_ref: bool = True,
    include_parent_child_ref: bool = True,
    include_root_child_ref: bool = True,
    group_by_resource: bool = False,
) -> str:
    """Convert a `dlt.Schema` to a a Graphviz DOT string and return its value.

    Args:
        schema: dlt schema to convert
        include_dlt_tables: If True, include data tables and internal dlt tables. This will influence table
            references and groups produced.
        include_internal_dlt_ref: If True, include references between tables `_dlt_version`, `_dlt_loads` and `_dlt_pipeline_state`
        include_parent_child_ref: If True, include references from `child._dlt_parent_id` to `parent._dlt_id`
        include_root_child_ref: If True, include references from `child._dlt_root_id` to `root._dlt_id`
        group_by_resource: If True, group tables by resource and create subclusters.

    Returns:
        A DOT string of the schema
    """

    # TODO use the cluster attribute to group tables by resource: https://www.graphviz.org/docs/clusters/
    graphviz_dot = _get_graph_header(schema["name"], include_dlt_tables)

    if group_by_resource:
        graphviz_dot = _add_table_clusters(
            schema, graphviz_dot, include_dlt_tables=include_dlt_tables
        )
    else:
        graphviz_dot = _add_tables(schema, graphviz_dot, include_dlt_tables=include_dlt_tables)

    graphviz_dot = _add_references(
        schema,
        graphviz_dot,
        include_dlt_tables=include_dlt_tables,
        include_internal_dlt_ref=include_internal_dlt_ref,
        include_parent_child_ref=include_parent_child_ref,
        include_root_child_ref=include_root_child_ref,
    )

    graphviz_dot += "}"
    return graphviz_dot


def _render_dot_with_html(dot: str) -> str:
    """Render the DOT string using an HTML file with required JS dependencies.

    HTML output automatically gets zoom and pan features
    This allows to render DOT without any Python or system dependencies. However,
    an internet connection is required to load the JS bundles.

    d3-graphviz is used because it's the smallest option (300Kb): https://github.com/magjac/d3-graphviz
    """

    HTML_TEMPLATE = """
<script src="//d3js.org/d3.v7.min.js"></script>
<script src="https://unpkg.com/@hpcc-js/wasm@2.20.0/dist/graphviz.umd.js"></script>
<script src="https://unpkg.com/d3-graphviz@5.6.0/build/d3-graphviz.js"></script>

<div id="graph" style="width:100%;height:100vh;display:flex;justify-content:center;align-items:center;"></div>
<script>
    d3.select("#graph")
      .graphviz({{fit: true}})
      .renderDot(
        `
        {dot}
        `
      );
</script>
"""
    return HTML_TEMPLATE.format(dot=dot)


def _render_with_graphviz(
    dot_source: graphviz.Source,
    path: Union[pathlib.Path, str],
    format_: Optional[str],
    save_dot_file: bool,
    render_kwargs: Optional[dict[str, Any]],
) -> pathlib.Path:
    """Resolve `path` and `format_` to render Graphviz object. Returns the output path.

    This handles quirks about graphviz path handling.
    """
    path = pathlib.Path(path).resolve()
    path.parent.mkdir(parents=True, exist_ok=True)

    # format is a string without `.`; it should be `"jpeg"`, not `".jpeg"`
    _, _, suffix = path.suffix.partition(".")

    if format_ is not None:
        graphviz_path_kwarg = path.with_suffix("") if suffix == format_ else path
    else:
        format_ = suffix
        # strip the `.suffix` because graphviz will append it during rendering
        graphviz_path_kwarg = path.with_suffix("")

    render_kwargs = {} if render_kwargs is None else render_kwargs
    # remove those kwargs if present because they would clash with `format_` and `save_dot_file`
    render_kwargs.pop("format", None)
    render_kwargs.pop("cleanup", None)

    output_path = dot_source.render(
        graphviz_path_kwarg,
        format=format_,
        cleanup=not save_dot_file,
        **render_kwargs,
    )

    return pathlib.Path(output_path)


def render_schema_with_graphviz(
    obj: Union[dlt.Schema, TStoredSchema, dlt.Pipeline, dlt.Dataset],
    *,
    path: Union[pathlib.Path, str],
    format_: Optional[str] = None,
    save_dot_file: bool = False,
    render_kwargs: Optional[dict[str, Any]] = None,
    include_dlt_tables: bool = True,
    include_internal_dlt_ref: bool = True,
    include_parent_child_ref: bool = True,
    include_root_child_ref: bool = True,
    group_by_resource: bool = False,
) -> pathlib.Path:
    """Render a visualization of the `dlt.Schema`.

    Args:
        obj: `dlt.Schema` or object with a `dlt.Schema` attached.
        path: file path for the rendered visualization. The render format is inferred from
            the file extension if `file_format` is unspecified.
        format: If specified, it will be used by the graphviz renderer and be appended to the `path` value.
        save_dot_file: If True, this function will output the content of the DOT string to a file. The DOT file
            allows to rerender the visualization. It can be useful to version it since it produces meaningful
            diffs. The DOT file is produced even if rendering fails, which can help debugging.
        render_kwargs: Dictionary of kwargs to pass to the Graphviz method `.render()`. This excludes `format`
            and `cleanup` which respectively match `format_` and `not save_dot_file`.
        include_dlt_tables: If True, include data tables and internal dlt tables. This will influence table
            references and groups produced.
        include_internal_dlt_ref: If True, include references between tables `_dlt_version`, `_dlt_loads` and `_dlt_pipeline_state`
        include_parent_child_ref: If True, include references from `child._dlt_parent_id` to `parent._dlt_id`
        include_root_child_ref: If True, include references from `child._dlt_root_id` to `root._dlt_id`
        group_by_resource: If True, group tables by resource into a subcluser.

    Returns:
        File path of the rendered dlt schema visualization.
    """
    try:
        import graphviz
    except ModuleNotFoundError:
        raise MissingDependencyException(
            "graphviz",
            ["graphviz"],
            "Install `graphviz` to be able to render schema as .png, .jpeg, .svg, etc.",
        )

    if isinstance(obj, dlt.Schema):
        stored_schema = obj.to_dict()
    # TODO better type check on `TStoredSchema`; can't check isinstance against TypedDict
    elif isinstance(obj, dict):
        stored_schema = obj
    elif isinstance(obj, dlt.Pipeline):
        stored_schema = obj.default_schema.to_dict()
    elif isinstance(obj, ReadableDBAPIDataset):  # TODO fix this check
        stored_schema = obj.schema.to_dict()
    else:
        raise TypeErrorWithKnownTypes(
            "obj", obj, ["dlt.Schema", "dlt.Pipeline", "dlt.Dataset", "TStoredSchema"]
        )

    dot_string = schema_to_graphviz(
        stored_schema,
        include_dlt_tables=include_dlt_tables,
        include_internal_dlt_ref=include_internal_dlt_ref,
        include_parent_child_ref=include_parent_child_ref,
        include_root_child_ref=include_root_child_ref,
        group_by_resource=group_by_resource,
    )

    output_path = _render_with_graphviz(
        dot_source=graphviz.Source(source=dot_string),
        path=path,
        format_=format_,
        save_dot_file=save_dot_file,
        render_kwargs=render_kwargs,
    )
    return output_path



from __future__ import annotations

from typing import (
    Optional,
    Iterator,
    Any,
    Sequence,
    AnyStr,
    Union,
    Tuple,
    List,
    Dict,
    Set,
    cast,
    TYPE_CHECKING,
)
from contextlib import contextmanager
from functools import wraps
import inspect
from pathlib import Path

from dlt.common.libs.sql_alchemy import sa
from sqlalchemy.engine import Connection
from sqlalchemy.exc import ResourceClosedError

from dlt.common.destination import DestinationCapabilitiesContext
from dlt.destinations.exceptions import (
    DatabaseUndefinedRelation,
    DatabaseTerminalException,
    DatabaseTransientException,
    LoadClientNotConnected,
    DatabaseException,
)
from dlt.common.destination.dataset import DBApiCursor
from dlt.common.typing import TFun

from dlt.destinations.typing import DBTransaction
from dlt.destinations.sql_client import SqlClientBase, raise_database_error
from dlt.destinations.impl.sqlalchemy.configuration import SqlalchemyCredentials
from dlt.destinations.impl.sqlalchemy.alter_table import MigrationMaker
from dlt.destinations.sql_client import DBApiCursorImpl

if TYPE_CHECKING:
    from dlt.destinations.impl.sqlalchemy.dialect import DialectCapabilities


class SqlaTransactionWrapper(DBTransaction):
    def __init__(self, sqla_transaction: sa.engine.Transaction) -> None:
        self.sqla_transaction = sqla_transaction

    def commit_transaction(self) -> None:
        if self.sqla_transaction.is_active:
            self.sqla_transaction.commit()

    def rollback_transaction(self) -> None:
        if self.sqla_transaction.is_active:
            self.sqla_transaction.rollback()


class SqlaDbApiCursor(DBApiCursorImpl):
    def __init__(self, curr: sa.engine.CursorResult) -> None:
        # Sqlalchemy CursorResult is *mostly* compatible with DB-API cursor
        self.native_cursor = curr  # type: ignore[assignment]
        curr.columns

        self._set_default_schema_columns()

    def _get_columns(self) -> List[str]:
        try:
            return list(cast(sa.engine.CursorResult, self.native_cursor).keys())
        except ResourceClosedError:
            # this happens if no rows are returned
            return []

    # @property
    # def description(self) -> Any:
    #     # Get the underlying driver's cursor description, this is mostly used in tests
    #     return self.native_cursor.cursor.description  # type: ignore[attr-defined]

    def execute(self, query: AnyStr, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError("execute not implemented")


class DbApiProps:
    # Only needed for some tests
    paramstyle = "named"


class SqlalchemyClient(SqlClientBase[Connection]):
    dbapi = DbApiProps  # type: ignore[assignment]
    migrations: Optional[MigrationMaker] = None  # lazy init as needed

    def __init__(
        self,
        dataset_name: str,
        staging_dataset_name: str,
        credentials: SqlalchemyCredentials,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        super().__init__(credentials.database, dataset_name, staging_dataset_name, capabilities)
        self.credentials = credentials
        self._dialect_caps: DialectCapabilities = capabilities.dialect_capabilities
        self._current_connection: Optional[Connection] = None
        self._current_transaction: Optional[SqlaTransactionWrapper] = None
        self.metadata = sa.MetaData()
        # Keep a list of datasets already attached on the current connection
        self._sqlite_attached_datasets: Set[str] = set()

    @property
    def engine(self) -> sa.engine.Engine:
        return self.credentials.engine

    @property
    def dialect(self) -> sa.engine.interfaces.Dialect:
        return self.engine.dialect

    @property
    def dialect_name(self) -> str:
        return self.dialect.name

    def open_connection(self) -> Connection:
        if self._current_connection is None:
            self._current_connection = self.credentials.managed_engine.borrow_conn()
            if self.dialect_name == "sqlite":
                self._sqlite_reattach_dataset_if_exists(self.dataset_name)
        return self._current_connection

    def close_connection(self) -> None:
        # rollback any transactions
        try:
            if self._in_transaction():
                self.rollback_transaction()
        except Exception:
            pass
        finally:
            self._current_transaction = None
        # detach databases
        try:
            if self.dialect_name == "sqlite":
                for dataset_name in list(self._sqlite_attached_datasets):
                    self._sqlite_detach_dataset(dataset_name)
        except Exception:
            # close internal connection. that will detach all datasets
            self._current_connection.connection.close()

        try:
            if self._current_connection is not None:
                self.credentials.managed_engine.return_conn(self._current_connection)
        finally:
            self._current_connection = None

    @property
    def native_connection(self) -> Connection:
        return self._current_connection

    def _in_transaction(self) -> bool:
        return (
            self._current_transaction is not None
            and self._current_transaction.sqla_transaction.is_active
        )

    @contextmanager
    @raise_database_error
    def begin_transaction(self) -> Iterator[DBTransaction]:
        trans = self._current_transaction = SqlaTransactionWrapper(self._current_connection.begin())
        try:
            yield trans
        except Exception:
            if self._in_transaction():
                self.rollback_transaction()
            raise
        else:
            if self._in_transaction():  # Transaction could be committed/rolled back before __exit__
                self.commit_transaction()
        finally:
            self._current_transaction = None

    @raise_database_error
    def commit_transaction(self) -> None:
        """Commits the current transaction."""
        self._current_transaction.commit_transaction()

    @raise_database_error
    def rollback_transaction(self) -> None:
        """Rolls back the current transaction."""
        self._current_transaction.rollback_transaction()

    @contextmanager
    @raise_database_error
    def _ensure_transaction(self) -> Iterator[DBTransaction]:
        """Context manager yielding either a new or the currently open transaction.
        New transaction will be committed/rolled back on exit.
        If the transaction is already open, finalization is handled by the top level context manager.
        """
        if self._in_transaction():
            yield self._current_transaction
            return
        with self.begin_transaction() as tx:
            yield tx

    def has_dataset(self) -> bool:
        with self._ensure_transaction():
            schema_names = self.engine.dialect.get_schema_names(self._current_connection)  # type: ignore[attr-defined]
        return self.dataset_name in schema_names

    def _sqlite_dataset_filename(self, dataset_name: str) -> str:
        current_file_path = Path(self.database_name)
        return str(
            current_file_path.parent
            / f"{current_file_path.stem}__{dataset_name}{current_file_path.suffix}"
        )

    def _sqlite_is_memory_db(self) -> bool:
        return SqlalchemyCredentials.is_memory_database(
            self.credentials.database, self.credentials.query
        )

    def _sqlite_reattach_dataset_if_exists(self, dataset_name: str) -> None:
        """Re-attach previously created databases for a new sqlite connection"""
        if self._sqlite_is_memory_db():
            return
        new_db_fn = self._sqlite_dataset_filename(dataset_name)
        if Path(new_db_fn).exists():
            self._sqlite_create_dataset(dataset_name)

    def _sqlite_create_dataset(self, dataset_name: str) -> None:
        """Mimic multiple schemas in sqlite using ATTACH DATABASE to
        attach a new database file to the current connection.
        """
        if not self._sqlite_is_memory_db():
            new_db_fn = self._sqlite_dataset_filename(dataset_name)

            if dataset_name != "main":  # main is the current file, it is always attached
                statement = "ATTACH DATABASE :fn AS :name"
                self.execute_sql(statement, fn=new_db_fn, name=dataset_name)
            # WAL mode is applied to all currently attached databases
            self.execute_sql("PRAGMA journal_mode=WAL")
            self.execute_sql("PRAGMA synchronous=NORMAL;")
        self._sqlite_attached_datasets.add(dataset_name)

    def _sqlite_drop_dataset(self, dataset_name: str) -> None:
        """Drop a dataset in sqlite by detaching the database file
        attached to the current connection.
        """
        # Get a list of attached databases and filenames
        rows = self.execute_sql("PRAGMA database_list")
        dbs = {row[1]: row[2] for row in rows}  # db_name: filename
        if dataset_name != "main":  # main is the default database, it cannot be detached
            self._sqlite_detach_dataset(dataset_name)

        fn = dbs[dataset_name]
        if not fn:  # It's a memory database, nothing to do
            return
        # Delete the database file
        Path(fn).unlink()

    def _sqlite_detach_dataset(self, dataset_name: str) -> None:
        statement = "DETACH DATABASE :name"
        self.execute_sql(statement, name=dataset_name)
        self._sqlite_attached_datasets.discard(dataset_name)

    @contextmanager
    def with_alternative_dataset_name(
        self, dataset_name: str
    ) -> Iterator[SqlClientBase[Connection]]:
        with super().with_alternative_dataset_name(dataset_name):
            if self.dialect_name == "sqlite" and dataset_name not in self._sqlite_attached_datasets:
                if not self.native_connection:
                    # opening connection attaches dataset
                    with self:
                        pass
                else:
                    self._sqlite_reattach_dataset_if_exists(dataset_name)
            yield self

    def create_dataset(self) -> None:
        if self.dialect_name == "sqlite":
            return self._sqlite_create_dataset(self.dataset_name)
        self.execute_sql(sa.schema.CreateSchema(self.dataset_name))

    def drop_dataset(self) -> None:
        if self.dialect_name == "sqlite":
            return self._sqlite_drop_dataset(self.dataset_name)
        try:
            self.execute_sql(sa.schema.DropSchema(self.dataset_name, cascade=True))
        except DatabaseException:  # Try again in case cascade is not supported
            self.execute_sql(sa.schema.DropSchema(self.dataset_name))

    def truncate_tables(self, *tables: str) -> None:
        # TODO: alchemy doesn't have a construct for TRUNCATE TABLE
        for table in tables:
            tbl = sa.Table(table, self.metadata, schema=self.dataset_name, keep_existing=True)
            self.execute_sql(tbl.delete())

    def drop_tables(self, *tables: str) -> None:
        for table in tables:
            tbl = sa.Table(table, self.metadata, schema=self.dataset_name, keep_existing=True)
            self.execute_sql(sa.schema.DropTable(tbl, if_exists=True))

    def execute_sql(
        self, sql: Union[AnyStr, sa.sql.Executable], *args: Any, **kwargs: Any
    ) -> Optional[Sequence[Sequence[Any]]]:
        with self.execute_query(sql, *args, **kwargs) as cursor:
            if cast(sa.engine.CursorResult, cursor).returns_rows:
                return cursor.fetchall()
            return None

    @contextmanager
    def execute_query(
        self, query: Union[AnyStr, sa.sql.Executable], *args: Any, **kwargs: Any
    ) -> Iterator[DBApiCursor]:
        if args and kwargs:
            raise ValueError("Cannot use both positional and keyword arguments")
        if isinstance(query, str):
            if args:
                # Sqlalchemy text supports :named paramstyle for all dialects
                query, kwargs = self._to_named_paramstyle(query, args)  # type: ignore[assignment]
                args = (kwargs,)
            query = sa.text(query)
        if kwargs:
            # sqla2 takes either a dict or list of dicts
            args = (kwargs,)
        with self._ensure_transaction():
            cur = self._current_connection.execute(query, *args)  # type: ignore[call-overload]
            try:
                yield SqlaDbApiCursor(cur)
            finally:
                cur.close()

    def get_existing_table(self, table_name: str) -> Optional[sa.Table]:
        """Get a table object from metadata if it exists"""
        key = self.dataset_name + "." + table_name
        return self.metadata.tables.get(key)  # type: ignore[no-any-return]

    def create_table(self, table_obj: sa.Table) -> None:
        with self._ensure_transaction():
            table_obj.create(self._current_connection)

    def make_qualified_table_name_path(
        self, table_name: Optional[str], quote: bool = True, casefold: bool = True
    ) -> List[str]:
        path: List[str] = []
        # no catalog for sqlalchemy
        if catalog_name := self.catalog_name(quote=quote, casefold=casefold):
            path.append(catalog_name)

        dataset_name = self.dataset_name
        if self.dialect.requires_name_normalize and casefold:  # type: ignore[attr-defined]
            dataset_name = str(self.dialect.normalize_name(dataset_name))  # type: ignore[func-returns-value]
        if quote:
            dataset_name = self.dialect.identifier_preparer.quote_identifier(dataset_name)  # type: ignore[attr-defined]
        path.append(dataset_name)
        if table_name:
            if self.dialect.requires_name_normalize and casefold:  # type: ignore[attr-defined]
                table_name = str(self.dialect.normalize_name(table_name))  # type: ignore[func-returns-value]
            if quote:
                table_name = self.dialect.identifier_preparer.quote_identifier(table_name)  # type: ignore[attr-defined]
            path.append(table_name)
        return path

    def alter_table_add_columns(self, columns: Sequence[sa.Column]) -> None:
        if not columns:
            return
        if self.migrations is None:
            self.migrations = MigrationMaker(self.dialect)
        for column in columns:
            self.migrations.add_column(column.table.name, column, schema=self.dataset_name)
        statements = self.migrations.consume_statements()
        for statement in statements:
            self.execute_sql(statement)

    def escape_column_name(
        self, column_name: str, quote: bool = True, casefold: bool = True
    ) -> str:
        if self.dialect.requires_name_normalize and casefold:  # type: ignore[attr-defined]
            column_name = str(self.dialect.normalize_name(column_name))  # type: ignore[func-returns-value]
        if quote:
            return self.dialect.identifier_preparer.quote_identifier(column_name)  # type: ignore[attr-defined,no-any-return]
        return column_name

    def reflect_table(
        self,
        table_name: str,
        metadata: Optional[sa.MetaData] = None,
        include_columns: Optional[Sequence[str]] = None,
    ) -> Optional[sa.Table]:
        """Reflect a table from the database and return the Table object"""
        if metadata is None:
            metadata = self.metadata
        try:
            with self._ensure_transaction():
                table = sa.Table(
                    table_name,
                    metadata,
                    autoload_with=self._current_connection,
                    resolve_fks=False,
                    schema=self.dataset_name,
                    include_columns=include_columns,
                    extend_existing=True,
                )
                # assert table is not None
                return table
        except DatabaseUndefinedRelation:
            return None

    def compare_storage_table(self, table_name: str) -> Tuple[sa.Table, List[sa.Column], bool]:
        """Reflect the table from database and compare it with the version already in metadata.
        Returns a 3 part tuple:
        - The current version of the table in metadata
        - List of columns that are missing from the storage table (all columns if it doesn't exist in storage)
        - boolean indicating whether the table exists in storage
        """
        existing = self.get_existing_table(table_name)
        assert existing is not None, "Table must be present in metadata"
        all_columns = list(existing.columns)
        all_column_names = [c.name for c in all_columns]
        tmp_metadata = sa.MetaData()
        reflected = self.reflect_table(
            table_name, include_columns=all_column_names, metadata=tmp_metadata
        )
        if reflected is None:
            missing_columns = all_columns
        else:
            missing_columns = [c for c in all_columns if c.name not in reflected.columns]
        return existing, missing_columns, reflected is not None

    def _make_database_exception(self, e: Exception) -> Exception:  # type: ignore[override]
        # NOTE: override uses self (not static) because it needs _dialect_caps
        from dlt.destinations.impl.sqlalchemy.dialect import GENERIC_TERMINAL_PATTERNS

        if isinstance(e, sa.exc.NoSuchTableError):
            return DatabaseUndefinedRelation(e)

        if isinstance(e, (sa.exc.ProgrammingError, sa.exc.OperationalError)):
            if self._dialect_caps.is_undefined_relation(e) is True:
                return DatabaseUndefinedRelation(e)
            msg = str(e).lower()
            for pat in GENERIC_TERMINAL_PATTERNS:
                if pat in msg:
                    return DatabaseTerminalException(e)
            return DatabaseTransientException(e)
        elif isinstance(e, sa.exc.IntegrityError):
            return DatabaseTerminalException(e)
        elif isinstance(e, sa.exc.DatabaseError):
            if self._dialect_caps.is_undefined_relation(e) is True:
                return DatabaseUndefinedRelation(e)
            return DatabaseTransientException(e)
        elif isinstance(e, sa.exc.SQLAlchemyError):
            return DatabaseTransientException(e)
        else:
            return e

    def _ensure_native_conn(self) -> None:
        if not self.native_connection:
            raise LoadClientNotConnected(type(self).__name__, self.dataset_name)



from dlt.common.destination.exceptions import DestinationTerminalException


class InvalidInMemoryQdrantCredentials(DestinationTerminalException):
    def __init__(self) -> None:
        super().__init__(
            "To use in-memory instance of qdrant, "
            "please instantiate it first and then pass to destination factory\n"
            '\nclient = QdrantClient(":memory:")\n'
            'dlt.pipeline(pipeline_name="...", destination=dlt.destinations.qdrant(client)'
        )



from typing import Any, AnyStr, List, Type, Optional, Protocol, Tuple, TypeVar, Generator


# native connection
TNativeConn = TypeVar("TNativeConn", bound=Any)

try:
    from pandas import DataFrame
except ImportError:
    DataFrame: Type[Any] = None  # type: ignore

try:
    from pyarrow import Table as ArrowTable
except ImportError:
    ArrowTable: Type[Any] = None  # type: ignore


class DBTransaction(Protocol):
    def commit_transaction(self) -> None: ...
    def rollback_transaction(self) -> None: ...


class DBApi(Protocol):
    threadsafety: int
    apilevel: str
    paramstyle: str



"""Fabric Warehouse destination for dlt"""

from dlt.common.destination.configuration import CsvFormatConfiguration
from dlt.destinations.impl.fabric.configuration import FabricCredentials, FabricClientConfiguration
from dlt.destinations.impl.fabric.factory import fabric

__all__ = [
    "fabric",
    "FabricCredentials",
    "FabricClientConfiguration",
    "CsvFormatConfiguration",
]



import re
from abc import ABC, abstractmethod

from typing import (
    Callable,
    ClassVar,
    List,
    Optional,
    Type,
    Union,
    Dict,
    Any,
    TypeVar,
    Generic,
)
from typing_extensions import TypeAlias
import inspect

import dlt
from dlt.common import logger
from dlt.common.configuration.specs.base_configuration import BaseConfiguration
from dlt.common.normalizers.naming import NamingConvention
from dlt.common.configuration import resolve_configuration, known_sections
from dlt.common.destination.capabilities import DestinationCapabilitiesContext
from dlt.common.destination.exceptions import (
    InvalidDestinationReference,
    UnknownDestinationModule,
)
from dlt.common.destination.client import DestinationClientConfiguration, JobClientBase
from dlt.common.runtime.run_context import get_plugin_modules
from dlt.common.schema.schema import Schema
from dlt.common.typing import is_subclass
from dlt.common.utils import get_full_callable_name, simple_repr, without_none
from dlt.common.reflection.ref import object_from_ref


TDestinationConfig = TypeVar("TDestinationConfig", bound="DestinationClientConfiguration")
TDestinationClient = TypeVar("TDestinationClient", bound="JobClientBase")
AnyDestination: TypeAlias = "Destination[DestinationClientConfiguration, JobClientBase]"
AnyDestination_CO: TypeAlias = "Destination[Any, Any]"


# TODO: type Destination properly
TDestinationReferenceArg = Union[str, AnyDestination_CO, Callable[..., AnyDestination_CO], None]


class Destination(ABC, Generic[TDestinationConfig, TDestinationClient]):
    """A destination factory that can be partially pre-configured
    with credentials and other config params.
    """

    config_params: Dict[str, Any]
    """Explicit config params, overriding any injected or default values."""
    caps_params: Dict[str, Any]
    """Explicit capabilities params, overriding any default values for this destination"""

    def __init__(self, **kwargs: Any) -> None:
        # Create initial unresolved destination config
        # Argument defaults are filtered out here because we only want arguments passed explicitly
        # to supersede config from the environment or pipeline args
        # __orig_base__ tells where the __init__ of interest is, in case class is derived
        sig = inspect.signature(getattr(self.__class__, "__orig_base__", self.__class__).__init__)
        params = sig.parameters

        # get available args
        spec = self.spec
        spec_fields = spec.get_resolvable_fields()
        caps_fields = DestinationCapabilitiesContext.get_resolvable_fields()

        # remove default kwargs
        kwargs = {k: v for k, v in kwargs.items() if k not in params or v != params[k].default}

        # warn on unknown params
        for k in list(kwargs):
            if k not in spec_fields and k not in caps_fields:
                logger.warning(
                    f"When initializing destination factory of type {self.destination_type},"
                    f" argument {k} is not a valid field in {spec.__name__} or destination"
                    " capabilities"
                )
                kwargs.pop(k)

        self.config_params = {k: v for k, v in kwargs.items() if k in spec_fields}
        self.caps_params = {k: v for k, v in kwargs.items() if k in caps_fields}

    @property
    @abstractmethod
    def spec(self) -> Type[TDestinationConfig]:
        """A spec of destination configuration that also contains destination credentials"""
        ...

    def capabilities(
        self, config: Optional[TDestinationConfig] = None, naming: Optional[NamingConvention] = None
    ) -> DestinationCapabilitiesContext:
        """Destination capabilities ie. supported loader file formats, identifier name lengths, naming conventions, escape function etc.
        Explicit caps arguments passed to the factory init and stored in `caps_params` are applied.

        If `config` is provided, it is used to adjust the capabilities, otherwise the explicit config composed just of `config_params` passed
          to factory init is applied
        If `naming` is provided, the case sensitivity and case folding are adjusted.
        """
        caps = self._raw_capabilities()
        caps.update(self.caps_params)
        # get explicit config if final config not passed
        if config is None:
            # create mock credentials to avoid credentials being resolved
            init_config = self.spec()
            init_config.update(self.config_params)
            if not init_config.credentials:
                credentials = self.spec.credentials_type(init_config)()
                credentials.__is_resolved__ = True
            else:
                credentials = init_config.credentials
                if isinstance(credentials, BaseConfiguration):
                    credentials = credentials.copy()
            config = self.spec(credentials=credentials)
            try:
                config = self.configuration(config, accept_partial=True)
            except Exception:
                # in rare cases partial may fail ie. when invalid native value is present
                # in that case we fallback to "empty" config
                pass
        caps = self.adjust_capabilities(caps, config, naming)
        # update again, explicit caps have prio
        caps.update(self.caps_params)
        return caps

    @abstractmethod
    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        """Returns raw capabilities, before being adjusted with naming convention and config"""
        ...

    def resolve_destination_name(self, destination_name: Optional[str]) -> str:
        """Resolves destination name by falling back to type-derived name if not explicitly provided."""
        return destination_name or self.to_name(self.destination_type)

    @property
    def destination_name(self) -> str:
        """The destination name will either be explicitly set while creating the destination or will be taken from the type"""
        return self.resolve_destination_name(self.config_params.get("destination_name"))

    @property
    def configured_name(self) -> str:
        """Configured destination name, None by default"""
        return self.config_params.get("destination_name")  # type: ignore[no-any-return]

    @property
    def destination_type(self) -> str:
        full_path = self.__class__.__module__ + "." + self.__class__.__qualname__
        return DestinationReference.normalize_type(full_path)

    @property
    def destination_description(self) -> str:
        return f"{self.destination_name} ({self.destination_type})"

    @property
    @abstractmethod
    def client_class(self) -> Type[TDestinationClient]:
        """A job client class responsible for starting and resuming load jobs"""
        ...

    def __repr__(self) -> str:
        # TODO: consider not showing default values of base SPEC
        #   consider showing physical location (when implemented) and
        #   props of the client ie. if it supports state, datasets or open tables
        kwargs = {**self.spec(), **self.config_params}
        return simple_repr(self.destination_type, **without_none(kwargs))

    def configuration(
        self, initial_config: TDestinationConfig, accept_partial: bool = False
    ) -> TDestinationConfig:
        """Get a fully resolved destination config from the initial config"""

        config = resolve_configuration(
            initial_config or self.spec(),
            sections=(known_sections.DESTINATION, self.destination_name),
            # already populated values will supersede resolved env config
            explicit_value=self.config_params,
            accept_partial=accept_partial,
        )
        return config

    def client(
        self, schema: Schema, initial_config: TDestinationConfig = None
    ) -> TDestinationClient:
        """Returns a configured instance of the destination's job client"""
        config = self.configuration(initial_config)
        caps = self.capabilities(config, schema.naming)
        # adjust naming for caps that dynamically set max length (ie. sql alchemy)
        if caps.max_identifier_length or caps.max_column_identifier_length:
            schema.naming.max_length = min(
                caps.max_identifier_length or caps.max_column_identifier_length,
                caps.max_column_identifier_length or caps.max_identifier_length,
            )
        return self.client_class(schema, config, caps)

    @classmethod
    def adjust_capabilities(
        cls,
        caps: DestinationCapabilitiesContext,
        config: TDestinationConfig,
        naming: Optional[NamingConvention],
    ) -> DestinationCapabilitiesContext:
        """Adjust the capabilities to match the case sensitivity as requested by naming convention."""
        # if naming not provided, skip the adjustment
        if not naming or not naming.is_case_sensitive:
            # all destinations are configured to be case insensitive so there's nothing to adjust
            return caps
        if not caps.has_case_sensitive_identifiers:
            if caps.casefold_identifier is str:
                logger.info(
                    f"Naming convention {naming.name()} is case sensitive but the destination does"
                    " not support case sensitive identifiers. Nevertheless identifier casing will"
                    " be preserved in the destination schema."
                )
            else:
                logger.warn(
                    f"Naming convention {naming.name()} is case sensitive but the destination does"
                    " not support case sensitive identifiers. Destination will case fold all the"
                    f" identifiers with {caps.casefold_identifier}"
                )
        else:
            # adjust case folding to store casefold identifiers in the schema
            if caps.casefold_identifier is not str:
                caps.casefold_identifier = str
                logger.info(
                    f"Enabling case sensitive identifiers for naming convention {naming.name()}"
                )
        return caps

    @staticmethod
    def to_name(ref: TDestinationReferenceArg) -> str:
        if ref is None:
            raise InvalidDestinationReference([])
        if isinstance(ref, str):
            return ref.rsplit(".", 1)[-1]
        if callable(ref):
            ref = ref()
        return ref.destination_name

    @classmethod
    def register(cls) -> None:
        """Registers this factory class under  __module__.__name__ of a Destination factory"""
        DestinationReference.register(cls, get_full_callable_name(cls))

    @classmethod
    def from_reference(
        cls,
        ref: TDestinationReferenceArg,
        credentials: Optional[Any] = None,
        destination_name: Optional[str] = None,
        environment: Optional[str] = None,
        **kwargs: Any,
    ) -> Optional[AnyDestination]:
        """Instantiate destination from a string reference or one of supported forms.
        This methods obtains a destination factory and then instantiates it passing
        the arguments after `ref`
        """
        # if we only get a name but no ref, we assume that the name is the destination_type
        if ref is None and destination_name is not None:
            ref = destination_name
        if ref is None:
            return None
        # evaluate callable returning Destination
        if callable(ref):
            ref = ref()
        if isinstance(ref, Destination):
            if credentials or destination_name or environment or kwargs:
                logger.warning(
                    "Cannot override credentials, destination_name, environment or kwargs when"
                    " passing a Destination instance, these values will be ignored."
                )
            return ref

        # If destination name is provided or ref is a module ref
        # don't attempt to resolve as named destination
        if destination_name or "." in ref:
            return DestinationReference.from_reference(
                ref, credentials, destination_name, environment, **kwargs
            )

        # First, try to resolve as a named destination with configured type
        destination_type: str = dlt.config.get(f"destination.{ref}.destination_type")
        if destination_type:
            try:
                return DestinationReference.from_reference(
                    ref=destination_type,
                    destination_name=ref,
                    credentials=credentials,
                    environment=environment,
                    **kwargs,
                )
            except Exception:
                pass

        # Then, try to resolve as a shorthand destination ref
        try:
            return DestinationReference.from_reference(
                ref, credentials, destination_name, environment, **kwargs
            )
        except UnknownDestinationModule as e:
            e.destination_type = destination_type
            e.named_dest_attempted = True
            raise


class DestinationReference:
    """A registry of destination factories with a set of method for finding and instantiating"""

    DESTINATIONS: ClassVar[Dict[str, Type[AnyDestination]]] = {}
    """A registry of all the destination factories"""

    @classmethod
    def register(cls, factory: Type[AnyDestination_CO], ref: str) -> None:
        """Registers `factory` class under `ref`. `ref`"""
        ref = cls.normalize_type(ref)
        if ref in cls.DESTINATIONS:
            logger.debug(
                f"A destination with ref {ref} is already registered and will be overwritten"
            )
        cls.DESTINATIONS[ref] = factory

    @classmethod
    def find(
        cls,
        ref: str,
        /,
        raise_exec_errors: bool = False,
        import_missing_modules: bool = False,
    ) -> Union[Type[AnyDestination], Callable[..., AnyDestination]]:
        """Finds or auto-imports destination factory that can be further called in order to instantiate it
        The ref can be a destination name or import path pointing to a destination class (e.g. `dlt.destinations.postgres`)
        You can control auto-import behavior:
         - `raise_exec_errors` - will re-raise code execution errors in imported modules
         - `import_missing_modules` - will ignore missing dependencies during import by substituting
         them with dummy modules. this should be only used to manipulate local dev environment

         NOTE: find returns a factory class or a callable that will create factory instance (for custom destinations)
         use `ensure_factory` to extract factory class from callable
         TODO: synthesize a __call__ on custom destination (`destination`) factory type so this distinction is not
         needed
        """
        refs = cls.to_fully_qualified_refs(ref)
        factory: Type[AnyDestination] = None
        if ref not in refs:
            refs = [ref] + refs

        for ref_ in refs:
            if factory := cls.DESTINATIONS.get(ref_):
                return factory

        def _typechecker(t_: Any) -> Any:
            # or destination type
            if is_subclass(t_, Destination):
                return t_
            # or callable that has factory and will return it
            assert callable(t_) and hasattr(t_, "_factory")
            return t_

        import_traces = []

        # no reference found, try to import default module
        if not factory:
            for possible_type in refs:
                if "." not in possible_type:
                    continue
                factory, trace = object_from_ref(
                    possible_type,
                    _typechecker,
                    raise_exec_errors=raise_exec_errors,
                    import_missing_modules=import_missing_modules,
                )
                if factory:
                    return factory
                import_traces.append(trace)

        raise UnknownDestinationModule(ref, refs, import_traces)

    @classmethod
    def ensure_factory(
        cls, ref_factory: Union[Type[AnyDestination], Callable[..., AnyDestination]]
    ) -> Type[AnyDestination]:
        """Extract factory type from a callable creating factory instance."""
        if is_subclass(ref_factory, Destination):
            return ref_factory  # type: ignore[return-value]
        return ref_factory._factory  # type: ignore[no-any-return,union-attr]

    @classmethod
    def from_reference(
        cls,
        ref: str,
        credentials: Optional[Any] = None,
        destination_name: Optional[str] = None,
        environment: Optional[str] = None,
        **kwargs: Any,
    ) -> Optional[AnyDestination]:
        """Instantiate destination from str reference.
        The ref can be a destination name or import path pointing to a destination class (e.g. `dlt.destinations.postgres`)
        This methods obtains a destination factory and then instantiates it passing the arguments after `ref`
        """
        if not isinstance(ref, str):
            raise InvalidDestinationReference(ref)

        factory = cls.find(ref)

        if credentials:
            kwargs["credentials"] = credentials
        if destination_name:
            kwargs["destination_name"] = destination_name
        if environment:
            kwargs["environment"] = environment
        return factory(**kwargs)

    @staticmethod
    def normalize_type(destination_type: str) -> str:
        """Normalizes destination type string into a canonical form. Assumes that type names without dots correspond to built in destinations."""
        if "." not in destination_type:
            destination_type = "dlt.destinations." + destination_type
        # the next two lines shorten the dlt internal destination paths to dlt.destinations.<destination_type>
        pattern = r"\.destinations\.impl\.[a-zA-Z_][.a-zA-Z0-9_]*\."
        replacement = ".destinations."
        destination_type = re.sub(pattern, replacement, destination_type)
        return destination_type

    @staticmethod
    def to_fully_qualified_refs(ref: str) -> List[str]:
        """Converts ref into fully qualified form, return one or more alternatives for shorthand notations.
        Run context is injected if needed. Following formats are recognized
        - name
        NOTE: the last component of destination type serves as destination name if not explicitly specified
        """
        ref_split = ref.split(".")
        ref_parts = len(ref_split)
        if ref_parts < 2:
            # context name is needed
            refs = []
            for ref_prefix in get_plugin_modules():
                if ref_prefix:
                    ref_prefix = f"{ref_prefix}.{known_sections.DESTINATIONS}"
                else:
                    ref_prefix = f"{known_sections.DESTINATIONS}"
                refs.append(f"{ref_prefix}.{ref}")
            return refs

        return []



from typing import TYPE_CHECKING

# NOTE this warning is only trigger when type-checking.
# Move it outside the type checking block when closer to deprecation date.
if not TYPE_CHECKING:
    from dlt.common.warnings import DltDeprecationWarning

    DltDeprecationWarning(
        """Content from this module was moved to `dlt._dataset.relation`, which is internal. \
        If you need to import `ReadableDBAPIRelation`, import `dlt.Relation` instead.""",
        since="1.16.0",
        expected_due="2.0.0",
    )

from dlt.dataset.relation import Relation

ReadableDBAPIRelation = Relation



from typing import Any, Dict, Sequence, Union, Final

from dlt.destinations.utils import get_resource_for_adapter
from dlt.extract import DltResource
from dlt.extract.items import TTableHintTemplate


PARTITION_HINT: Final[str] = "x-athena-partition"


class PartitionTransformation:
    template: str
    """Template string of the transformation including column name placeholder. E.g. `bucket(16, {column_name})`"""
    column_name: str
    """Column name to apply the transformation to"""

    def __init__(self, template: str, column_name: str) -> None:
        self.template = template
        self.column_name = column_name


class athena_partition:
    """Helper class to generate iceberg partition transformations

    E.g. `athena_partition.bucket(16, "id")` will return a transformation with template `bucket(16, {column_name})`
    This can be correctly rendered by the athena loader with escaped column name.
    """

    @staticmethod
    def year(column_name: str) -> PartitionTransformation:
        """Partition by year part of a date or timestamp column."""
        return PartitionTransformation("year({column_name})", column_name)

    @staticmethod
    def month(column_name: str) -> PartitionTransformation:
        """Partition by month part of a date or timestamp column."""
        return PartitionTransformation("month({column_name})", column_name)

    @staticmethod
    def day(column_name: str) -> PartitionTransformation:
        """Partition by day part of a date or timestamp column."""
        return PartitionTransformation("day({column_name})", column_name)

    @staticmethod
    def hour(column_name: str) -> PartitionTransformation:
        """Partition by hour part of a date or timestamp column."""
        return PartitionTransformation("hour({column_name})", column_name)

    @staticmethod
    def bucket(n: int, column_name: str) -> PartitionTransformation:
        """Partition by hashed value to n buckets."""
        return PartitionTransformation(f"bucket({n}, {{column_name}})", column_name)

    @staticmethod
    def truncate(length: int, column_name: str) -> PartitionTransformation:
        """Partition by value truncated to length."""
        return PartitionTransformation(f"truncate({length}, {{column_name}})", column_name)


def athena_adapter(
    data: Any,
    partition: Union[
        str, PartitionTransformation, Sequence[Union[str, PartitionTransformation]]
    ] = None,
) -> DltResource:
    """
    Prepares data for loading into Athena

    Args:
        data: The data to be transformed.
            This can be raw data or an instance of DltResource.
            If raw data is provided, the function will wrap it into a `DltResource` object.
        partition: Column name(s) or instances of `PartitionTransformation` to partition the table by.
            To use a transformation it's best to use the methods of the helper class `athena_partition`
            to generate correctly escaped SQL in the loader.

    Returns:
        A `DltResource` object that is ready to be loaded into Athena.

    Raises:
        ValueError: If any hint is invalid or none are specified.

    Examples:
        >>> data = [{"name": "Marcel", "department": "Engineering", "date_hired": "2024-01-30"}]
        >>> athena_adapter(data, partition=["department", athena_partition.year("date_hired"), athena_partition.bucket(8, "name")])
        [DltResource with hints applied]
    """
    resource = get_resource_for_adapter(data)
    additional_table_hints: Dict[str, TTableHintTemplate[Any]] = {}

    if partition:
        if isinstance(partition, str) or not isinstance(partition, Sequence):
            partition = [partition]

        # Partition hint is `{column_name: template}`, e.g. `{"department": "{column_name}", "date_hired": "year({column_name})"}`
        # Use one dict for all hints instead of storing on column so order is preserved
        partition_hint: Dict[str, str] = {}

        for item in partition:
            if isinstance(item, PartitionTransformation):
                # Client will generate the final SQL string with escaped column name injected
                partition_hint[item.column_name] = item.template
            else:
                # Item is the column name
                partition_hint[item] = "{column_name}"

        additional_table_hints[PARTITION_HINT] = partition_hint

    if additional_table_hints:
        resource.apply_hints(additional_table_hints=additional_table_hints)
    else:
        raise ValueError("A value for `partition` must be specified.")
    return resource



import os
import itertools
from typing import List, Dict, NamedTuple, Sequence, Set, Optional, Callable
from concurrent.futures import Future, Executor

from dlt.common import logger
from dlt.common.metrics import DataWriterMetrics
from dlt.common.runtime.signals import sleep
from dlt.common.configuration import with_config, known_sections
from dlt.common.configuration.accessors import config
from dlt.common.data_writers.writers import EMPTY_DATA_WRITER_METRICS
from dlt.common.runners import TRunMetrics, Runnable, NullExecutor
from dlt.common.runtime import signals
from dlt.common.runtime.collector import Collector, NULL_COLLECTOR
from dlt.common.schema.typing import TSchemaUpdate, TStoredSchema, TTableSchema
from dlt.common.schema.utils import merge_schema_updates
from dlt.common.storages import (
    NormalizeStorage,
    SchemaStorage,
    LoadStorage,
    ParsedLoadJobFileName,
)
from dlt.common.schema import Schema
from dlt.common.schema.exceptions import CannotCoerceColumnException
from dlt.common.pipeline import (
    NormalizeInfo,
    NormalizeMetrics,
    SupportsPipeline,
    WithStepInfo,
)
from dlt.common.storages.exceptions import LoadPackageNotFound
from dlt.common.storages.load_package import LoadPackageInfo

from dlt.normalize.configuration import NormalizeConfiguration
from dlt.normalize.exceptions import NormalizeJobFailed
from dlt.normalize.worker import w_normalize_files, group_worker_files, TWorkerRV
from dlt.normalize.validate import (
    validate_and_update_schema,
    verify_normalized_table,
    reconcile_null_only_columns,
)


class SubmitRV(NamedTuple):
    schema_updates: List[TSchemaUpdate]
    file_metrics: List[DataWriterMetrics]
    pending_exc: BaseException
    null_only_columns: Dict[str, Set[str]]


# normalize worker wrapping function signature
TMapFuncType = Callable[
    [Schema, str, Sequence[str]], SubmitRV
]  # input parameters: (schema name, load_id, list of files to process)


def remove_files_from_metrics(file_metrics: List[DataWriterMetrics]) -> None:
    """Deletes files in a list of metrics"""
    for metrics in file_metrics:
        os.remove(metrics.file_path)


class Normalize(Runnable[Executor], WithStepInfo[NormalizeMetrics, NormalizeInfo]):
    pool: Executor

    @with_config(spec=NormalizeConfiguration, sections=(known_sections.NORMALIZE,))
    def __init__(
        self,
        collector: Collector = NULL_COLLECTOR,
        schema_storage: SchemaStorage = None,
        config: NormalizeConfiguration = config.value,
    ) -> None:
        self.config = config
        self.collector = collector
        self.normalize_storage: NormalizeStorage = None
        self.pool = NullExecutor()
        self.load_storage: LoadStorage = None
        self.schema_storage: SchemaStorage = None

        # setup storages
        self.create_storages()
        # create schema storage with give type
        self.schema_storage = schema_storage or SchemaStorage(
            self.config._schema_storage_config, makedirs=True
        )
        super().__init__()

    def create_storages(self) -> None:
        # pass initial normalize storage config embedded in normalize config
        self.normalize_storage = NormalizeStorage(
            True, config=self.config._normalize_storage_config
        )
        # normalize saves in preferred format but can read all supported formats
        self.load_storage = LoadStorage(
            True,
            LoadStorage.ALL_SUPPORTED_FILE_FORMATS,
            config=self.config._load_storage_config,
        )

    def map_parallel(self, schema: Schema, load_id: str, files: Sequence[str]) -> SubmitRV:
        workers: int = getattr(self.pool, "_max_workers", 1)
        # group files to process into as many groups as there are workers. prefer to send same tables
        # to the same worker
        chunk_files = group_worker_files(files, workers)
        schema_dict: TStoredSchema = schema.to_dict()
        param_chunk = [
            (
                self.config,
                self.normalize_storage.config,
                self.load_storage.config,
                schema_dict,
                load_id,
                files,
            )
            for files in chunk_files
        ]
        # return stats
        summary = TWorkerRV([], [], {})
        # push all tasks to queue
        tasks = [(self.pool.submit(w_normalize_files, *params), params) for params in param_chunk]
        pending_exc: BaseException = None
        logger.info(f"Filled {len(tasks)} out of {len(files)} files")

        while len(tasks) > 0:
            sleep(0.3)
            # operate on copy of the list
            for task in list(tasks):
                pending, params = task
                if pending.done():
                    exc = pending.exception()
                    if exc is not None:
                        # collect metrics from NormalizeJobFailed (if any), then cancel others and propagate
                        if isinstance(exc, NormalizeJobFailed):
                            summary.file_metrics.extend(exc.writer_metrics)
                        logger.warning("Received exception from worker: %s" % str(exc))
                        if not pending_exc:
                            # store to raise when pool is drained
                            pending_exc = exc
                            # cancel workers on any error
                            logger.warning(f"Cancelling package {load_id}")
                            self.load_storage.new_packages.cancel(load_id)
                    else:
                        result: TWorkerRV = pending.result()
                        try:
                            # gather schema from all manifests, validate consistency and combine
                            validate_and_update_schema(schema, result[0])
                            summary.schema_updates.extend(result.schema_updates)
                            summary.file_metrics.extend(result.file_metrics)
                            for tn, cols in result.null_only_columns.items():
                                summary.null_only_columns.setdefault(tn, set()).update(cols)
                            # update metrics
                            self.collector.update("Files", len(result.file_metrics))
                            self.collector.update(
                                "Items",
                                sum(result.file_metrics, EMPTY_DATA_WRITER_METRICS).items_count,
                            )
                        except CannotCoerceColumnException as exc:
                            # schema conflicts resulting from parallel executing
                            logger.warning(
                                f"Parallel schema update conflict, retrying task ({str(exc)}"
                            )
                            # delete all files produced by the task
                            remove_files_from_metrics(result.file_metrics)
                            # schedule the task again with the schema updated by other workers
                            schema_dict = schema.to_dict()
                            # TODO: it's time for a named tuple
                            params = params[:3] + (schema_dict,) + params[4:]
                            retry_pending: Future[TWorkerRV] = self.pool.submit(
                                w_normalize_files, *params
                            )
                            tasks.append((retry_pending, params))
                    # remove finished tasks
                    tasks.remove(task)
                    logger.debug(f"{len(tasks)} tasks still remaining for {load_id}...")
            # raise on signal
            if signals.was_signal_received() and not pending_exc:
                pending_exc = signals.exception_for_signal()
                logger.warning(f"Cancelling package {load_id} due to signal")
                self.load_storage.new_packages.cancel(load_id)

        return SubmitRV(
            summary.schema_updates, summary.file_metrics, pending_exc, summary.null_only_columns
        )

    def map_single(self, schema: Schema, load_id: str, files: Sequence[str]) -> SubmitRV:
        pending_exc: Exception = None
        try:
            result = w_normalize_files(
                self.config,
                self.normalize_storage.config,
                self.load_storage.config,
                schema.to_dict(),
                load_id,
                files,
            )
            validate_and_update_schema(schema, result.schema_updates)
            self.collector.update("Files", len(result.file_metrics))
            self.collector.update(
                "Items", sum(result.file_metrics, EMPTY_DATA_WRITER_METRICS).items_count
            )
        except NormalizeJobFailed as job_failed_ex:
            pending_exc = job_failed_ex
            result = TWorkerRV(None, job_failed_ex.writer_metrics, {})

        return SubmitRV(
            result.schema_updates, result.file_metrics, pending_exc, result.null_only_columns
        )

    def clean_x_normalizer(self, load_id: str, table_name: str, table_schema: TTableSchema) -> None:
        x_normalizer = table_schema.setdefault("x-normalizer", {})
        # drop evolve once for all tables that seen data
        x_normalizer.pop("evolve-columns-once", None)
        # mark that table have seen data only if there was data
        if "seen-data" not in x_normalizer:
            logger.info(
                f"Table {table_name} has seen data for the first time with load id {load_id}"
            )
            x_normalizer["seen-data"] = True

    def spool_files(
        self, load_id: str, schema: Schema, map_f: TMapFuncType, files: Sequence[str]
    ) -> None:
        # process files in parallel or in single thread, depending on map_f
        schema_updates, writer_metrics, pending_exc, null_only_columns = map_f(
            schema, load_id, files
        )
        # compute metrics
        job_metrics = {ParsedLoadJobFileName.parse(m.file_path): m for m in writer_metrics}
        table_metrics: Dict[str, DataWriterMetrics] = {
            table_name: sum(map(lambda pair: pair[1], metrics), EMPTY_DATA_WRITER_METRICS)
            for table_name, metrics in itertools.groupby(
                job_metrics.items(), lambda pair: pair[0].table_name
            )
        }
        self._step_info_update_metrics(
            load_id,
            {
                "started_at": None,
                "finished_at": None,
                "job_metrics": {job.job_id(): metrics for job, metrics in job_metrics.items()},
                "table_metrics": table_metrics,
            },
        )
        # raise pending exception after computing metrics
        if pending_exc:
            raise pending_exc

        # reconcile null-only columns against schema
        reconciled_null_cols = reconcile_null_only_columns(schema, null_only_columns)

        # update normalizer specific info
        for table_name in table_metrics:
            table = schema.tables[table_name]
            verify_normalized_table(
                schema,
                table,
                self.config.destination_capabilities,
                reconciled_null_cols.get(table_name),
            )
            self.clean_x_normalizer(load_id, table_name, table)
        # schema is updated, save it to schema volume
        if schema.is_modified:
            logger.info(
                f"Saving schema {schema.name} with version {schema.stored_version}:{schema.version}"
            )
            self.schema_storage.save_schema(schema)
        else:
            logger.info(
                f"Schema {schema.name} with version {schema.version} was not modified. Save skipped"
            )
        # TODO: do not use schema_updates to generate a diff. compare the schema in the storage
        # to the schema being saved and create unified diff. schema may change in several places (also extraction)
        # save schema new package
        self.load_storage.new_packages.save_schema(load_id, schema)
        # save schema updates even if empty
        self.load_storage.new_packages.save_schema_updates(
            load_id, merge_schema_updates(schema_updates)
        )
        # files must be renamed and deleted together so do not attempt that when process is about to be terminated
        signals.raise_if_signalled()
        logger.info("Committing storage, do not kill this process")
        # rename temp folder to processing
        self.load_storage.commit_new_load_package(load_id)
        # delete item files to complete commit
        self.normalize_storage.extracted_packages.delete_package(load_id)
        # log and update metrics
        logger.info(f"Extracted package {load_id} processed")
        self._step_info_complete_load_id(load_id)

    def spool_schema_files(self, load_id: str, schema: Schema, files: Sequence[str]) -> str:
        # delete existing folder for the case that this is a retry
        self.load_storage.new_packages.delete_package(load_id, not_exists_ok=True)
        # normalized files will go here before being atomically renamed
        self.load_storage.import_extracted_package(
            load_id, self.normalize_storage.extracted_packages
        )
        logger.info(f"Created new load package {load_id} on loading volume with {len(files)} files")
        # get number of workers with default == 1 if not set (ie. NullExecutor)
        workers: int = getattr(self.pool, "_max_workers", 1)
        map_f: TMapFuncType = self.map_parallel if workers > 1 else self.map_single
        self.spool_files(load_id, schema.clone(update_normalizers=True), map_f, files)

        return load_id

    def run(self, pool: Optional[Executor]) -> TRunMetrics:
        # keep the pool in class instance
        self.pool = pool or NullExecutor()
        logger.info("Running file normalizing")
        # list all load packages in extracted folder
        load_ids = self.normalize_storage.extracted_packages.list_packages()
        logger.info(f"Found {len(load_ids)} load packages")
        if len(load_ids) == 0:
            return TRunMetrics(True, 0)
        for load_id in load_ids:
            # read schema from package
            schema = self.normalize_storage.extracted_packages.load_schema(load_id)
            # prefer schema from schema storage if it exists
            try:
                # use live schema instance via getter if on live storage, it will also do import
                # schema as live schemas are committed before calling normalize
                storage_schema = self.schema_storage[schema.name]
                if schema.stored_version_hash != storage_schema.stored_version_hash:
                    logger.warning(
                        f"When normalizing package {load_id} with schema {schema.name}: the storage"
                        f" schema hash {storage_schema.stored_version_hash} is different from"
                        f" extract package schema hash {schema.stored_version_hash}. Storage schema"
                        " was used."
                    )
                schema = storage_schema
            except FileNotFoundError:
                pass
            # read all files to normalize placed as new jobs
            schema_files = self.normalize_storage.extracted_packages.list_new_jobs(load_id)
            logger.info(
                f"Found {len(schema_files)} files in schema {schema.name} load_id {load_id}"
            )
            if len(schema_files) == 0:
                # delete empty package
                self.normalize_storage.extracted_packages.delete_package(load_id)
                logger.info(f"Empty package {load_id} processed")
                continue
            with self.collector(f"Normalize {schema.name} in {load_id}"):
                self.collector.update("Files", 0, len(schema_files))
                self.collector.update("Items", 0)
                # self.verify_package(load_id, schema, schema_files)
                self._step_info_start_load_id(load_id)
                self.spool_schema_files(load_id, schema, schema_files)

        # return info on still pending packages (if extractor saved something in the meantime)
        return TRunMetrics(False, len(self.normalize_storage.extracted_packages.list_packages()))

    # def verify_package(self, load_id, schema: Schema, schema_files: Sequence[str]) -> None:
    #     """Verifies package schema and jobs against destination capabilities"""
    #     # get all tables in schema files
    #     table_names = set(ParsedLoadJobFileName.parse(job).table_name for job in schema_files)

    def get_load_package_info(self, load_id: str) -> LoadPackageInfo:
        """Returns information on extracted/normalized/completed package with given load_id, all jobs and their statuses."""
        try:
            return self.load_storage.get_load_package_info(load_id)
        except LoadPackageNotFound:
            # combine new extracted package and new load package that will be discarded
            new_package = self.load_storage.new_packages.get_load_package_info(load_id)
            extracted_package = self.normalize_storage.extracted_packages.get_load_package_info(
                load_id
            )
            extracted_package.jobs["completed_jobs"] = new_package.jobs["new_jobs"]
            return extracted_package

    def get_step_info(
        self,
        pipeline: SupportsPipeline,
    ) -> NormalizeInfo:
        load_ids = list(self._load_id_metrics.keys())
        load_packages: List[LoadPackageInfo] = []
        metrics: Dict[str, List[NormalizeMetrics]] = {}
        for load_id in self._load_id_metrics.keys():
            load_package = self.get_load_package_info(load_id)
            load_packages.append(load_package)
            metrics[load_id] = self._step_info_metrics(load_id)
        return NormalizeInfo(pipeline, metrics, load_ids, load_packages, pipeline.first_run)



from dlt.common.schema.typing import (
    TSchemaContractDict,
    TSchemaUpdate,
    TSchemaTables,
    TTableSchema,
    TStoredSchema,
    TTableSchemaColumns,
    TColumnHint,
    TColumnSchema,
    TColumnSchemaBase,
)
from dlt.common.schema.typing import COLUMN_HINTS
from dlt.common.schema.schema import Schema, DEFAULT_SCHEMA_CONTRACT_MODE
from dlt.common.schema.exceptions import DataValidationError
from dlt.common.schema.utils import verify_schema_hash

__all__ = [
    "TSchemaUpdate",
    "TSchemaTables",
    "TTableSchema",
    "TStoredSchema",
    "TTableSchemaColumns",
    "TColumnHint",
    "TColumnSchema",
    "TColumnSchemaBase",
    "COLUMN_HINTS",
    "Schema",
    "verify_schema_hash",
    "TSchemaContractDict",
    "DEFAULT_SCHEMA_CONTRACT_MODE",
    "DataValidationError",
]



from typing import Any, Callable, List, Literal, Optional, Sequence, TypeVar, Union

from dlt.common.typing import TSortOrder, TTableHintTemplate, TColumnNames, TypedDict

TCursorValue = TypeVar("TCursorValue", bound=Any)
LastValueFunc = Callable[[Sequence[TCursorValue]], Any]
OnCursorValueMissing = Literal["raise", "include", "exclude"]

TIncrementalRange = Literal["open", "closed"]


class IncrementalColumnState(TypedDict):
    initial_value: Optional[Any]
    last_value: Optional[Any]
    unique_hashes: List[str]


class IncrementalArgs(TypedDict, total=False):
    cursor_path: str
    initial_value: Optional[Any]
    last_value_func: Optional[Union[LastValueFunc[str], Literal["min", "max"]]]
    """Last value callable or name of built in function"""
    primary_key: Optional[TTableHintTemplate[TColumnNames]]
    end_value: Optional[Any]
    row_order: Optional[TSortOrder]
    allow_external_schedulers: Optional[bool]
    lag: Optional[Union[float, int]]
    on_cursor_value_missing: Optional[OnCursorValueMissing]
    range_start: Optional[TIncrementalRange]
    range_end: Optional[TIncrementalRange]
